from __future__ import print_function
import SCons.Script
import SCons.Util
import collections
import copy
import sys

# Formats message, prints to stderr, and exits with error.
def Die(env, fmt, *args):
    print('error: {}\n'.format(fmt.format(*args).strip()), file=sys.stderr)
    SCons.Script.Exit(1)

# Returns a deep clone of env.
# Workaround for buggy behavior of env.Clone() in recent scons.
def DeepClone(env):
    cloned_env = env.Clone()

    # fixup after env.Clone()
    for key in env.Dictionary():
        if not cloned_env[key] is env[key]:
            # already copied
            continue
        if callable(cloned_env[key]) \
          or isinstance(cloned_env[key], SCons.Util.Unbuffered):
            # not copyable
            continue
        if isinstance(cloned_env[key], (int, float, complex, bool, str, tuple)) \
          or cloned_env[key] is None:
            # immutable
            continue

        try:
            # try to do deep copy
            cloned_value = copy.deepcopy(env[key])
            cloned_env[key] = cloned_value
        except:
            pass

    return cloned_env

# env.MergeFrom(other_env) merges configuration from other_env into env.
def MergeFrom(dst_env, src_env, exclude=[]):
    for key, src_val in src_env.Dictionary().items():
        if key in exclude:
            continue

        if isinstance(src_val, SCons.Util.CLVar) \
          or isinstance(src_val, collections.deque) \
          or isinstance(src_val, list):
            if key in dst_env.Dictionary():
                for item in src_val:
                    if item in dst_env[key]:
                        dst_env[key].remove(item)
            if 'FLAGS' in key:
                dst_env.Append(**{key: src_val})
            else:
                dst_env.AppendUnique(**{key: src_val})
        else:
            if key not in dst_env.Dictionary():
                dst_env[key] = src_val

def init(env):
    env.AddMethod(Die, 'Die')
    env.AddMethod(DeepClone, 'DeepClone')
    env.AddMethod(MergeFrom, 'MergeFrom')



from __future__ import print_function

import sys
import os
import os.path
import fnmatch
import json
import shlex

try:
    import fcntl

    def flock(fp):
        fcntl.flock(fp.fileno(), fcntl.LOCK_EX)

    def funlock(fp):
        fcntl.flock(fp.fileno(), fcntl.LOCK_UN)
except:
    import msvcrt

    def flock(fp):
        fp.seek(0)
        msvcrt.locking(fp.fileno(), msvcrt.LK_LOCK, 1)

    def funlock(fp):
        fp.seek(0)
        msvcrt.locking(fp.fileno(), msvcrt.LK_UNLCK, 1)

if len(sys.argv) < 5:
   print("usage: clangdb.py ROOT_DIR BUILD_DIR COMPILER COMPILER_ARGS...",
         file=sys.stderr)
   exit(1)

root_dir = os.path.abspath(sys.argv[1])
build_dir = os.path.abspath(sys.argv[2])
compiler = sys.argv[3]
compiler_args = sys.argv[4:]

source_file = None

for arg in compiler_args:
    for pattern in ['*.c', '*.cpp']:
        if fnmatch.fnmatch(arg, pattern):
            source_file = os.path.join(root_dir, arg)
            break
    if source_file:
        break

if source_file:
    db_path = os.path.join(build_dir, "compile_commands.json")

    cmd = {
        "command": "{} {}".format(compiler, ' '.join(compiler_args)),
        "directory": root_dir,
        "file": source_file,
    }

    try:
        with open(db_path, 'a+') as fp:
            try:
                flock(fp)

                try:
                    fp.seek(0)
                    db = json.loads(fp.read())
                    db[:]
                except:
                    db = []

                for index, item in enumerate(db):
                    if item["file"] == source_file:
                        db[index] = cmd
                        break
                else:
                    db.append(cmd)

                fp.seek(0)
                fp.truncate()
                fp.write(json.dumps(db, indent=2))
            finally:
                funlock(fp)
    except:
        e = sys.exc_info()[1]
        print("error: unable to write clangdb to " + db_path, file=sys.stderr)
        print(str(e), file=sys.stderr)

cmd = shlex.split(compiler) + compiler_args
os.execvp(cmd[0], cmd)



import SCons.SConf
import hashlib
import os
import os.path
import re
import textwrap

# Python 2 compatibility.
try:
    from shlex import quote
except:
    from pipes import quote

def _run_prog(context, src, suffix):
    # Workaround for a SCons bug.
    # RunProg uses a global incrementing counter for temporary .c file names. The
    # file name depends on the number of invocations of that function, but not on
    # the file contents. When the user subsequently invokes scons with different
    # options, the sequence of file contents passed to RunProg may vary. However,
    # RunProg may incorrectly use cached results from a previous run saved for
    # different file contents but the same invocation number. To prevent this, we
    # monkey patch its global counter with a hashsum of the file contents.
    # The workaround is needed only for older versions of SCons, where
    # _ac_build_counter was an integer.
    try:
        if type(SCons.SConf._ac_build_counter) is int:
            SCons.SConf._ac_build_counter = int(hashlib.md5(src.encode()).hexdigest(), 16)
    except:
        pass
    return context.RunProg(src, suffix)

def _get_llvm_dir(version):
    def macos_dirs():
        return [
        '/Applications/Xcode.app/Contents/Developer/Toolchains/XcodeDefault.xctoolchain/usr/bin',
        '/Library/Developer/CommandLineTools/usr/bin',
        ]

    def linux_dirs():
        suffixes = []
        for n in [3, 2, 1]:
            v = '.'.join(map(str, version[:n]))
            suffixes += [
                '-' + v,
                '/' + v,
            ]
        suffixes += ['']
        ret = []
        for s in suffixes:
            ret.append('/usr/lib/llvm{}/bin'.format(s))
        return ret

    for llvmdir in macos_dirs() + linux_dirs():
        if os.path.isdir(llvmdir):
            return llvmdir

# Gets `PKG_CONFIG_PATH` from environment `env`, tries to append `add_prefix`/lib/pkgconfig, and
# returns the result.
# Tolerant to absense of any or both of those two components. If both are absent, returns an
# empty string.
# Doesn't perform quoting.
def _compose_pkg_config_path(env, add_prefix):
    pkg_config_path = env.get('PKG_CONFIG_PATH', '')
    if add_prefix:
        # set user-defined $(add_prefix)/lib/pkgconfig path for pkg-config if needed (lowest
        # priority); otherwise use original PKG_CONFIG_PATH is propagated explicitly
        add_pkg_config_path = os.path.join(add_prefix, 'lib', 'pkgconfig')
        if os.path.isdir(add_pkg_config_path):
            if pkg_config_path:
                pkg_config_path += ':'
            pkg_config_path += add_pkg_config_path
    return pkg_config_path

# Workaround for brew / pkg-config weirdeness
def _fix_brew_libpath(libpath):
    if not '/Cellar/' in libpath:
        return libpath # not in brew (quick check)

    if libpath.endswith('/lib'):
        return libpath # looks good

    brew_prefix = env.GetCommandOutput('brew --prefix')
    if not brew_prefix:
        return libpath # can't help

    if not libpath.startswith(brew_prefix+'/Cellar/'):
        return libpath # not in brew

    if not os.path.isdir(libpath+'/lib'):
        return libpath # can't help

    return libpath+'/lib'

def CheckLibWithHeaderExt(context, libs, headers, language, expr='1', run=True):
    if not isinstance(headers, list):
        headers = [headers]

    if not isinstance(libs, list):
        libs = [libs]

    name = libs[0]
    libs = [l for l in libs if not l in context.env['LIBS']]

    suffix = '.{}'.format(language.lower())
    includes = '\n'.join(['#include <{}>'.format(h) for h in ['stdio.h'] + headers])
    src = textwrap.dedent("""
        {}

        int main() {{
            printf("%d\\n", (int)({}));
            return 0;
        }}
    """).format(includes, expr)

    context.Message("Checking for {} library {} ... ".format(language.upper(), name))

    if run:
        err, out = _run_prog(context, src, suffix)
        if out.strip() == '0':
            err = True
    else:
        err = context.CompileProg(src, suffix)

    if not err:
        context.Result('yes')
        context.env.Append(LIBS=libs)
        return True
    else:
        context.Result('no')
        return False

def CheckProg(context, prog):
    context.Message("Checking for executable {} ... ".format(prog))

    path = context.env.Which(prog)
    if path:
        context.Result(path[0])
        return True
    else:
        context.Result('not found')
        return False

def CheckCanRunProgs(context):
    context.Message("Checking whether we can run compiled executables ... ")

    src = textwrap.dedent("""
        int main() {
            return 0;
        }
    """)

    err, out = _run_prog(context, src, '.c')

    if not err:
        context.Result('yes')
        return True
    else:
        context.Result('no')
        return False

def CheckCompilerOptionSupported(context, opt, language):
    context.Message("Checking whether {} compiler supports {} ... ".format(language.upper(), opt))

    ext = '.' + language.lower()
    src = "int main() { return 0; }"

    orig_env = context.env

    context.env = orig_env.DeepClone()
    context.env.Append(**{language.upper()+'FLAGS': [opt]})

    err = context.CompileProg(src, ext)

    context.env = orig_env

    if not err:
        context.Result('yes')
        return True
    else:
        context.Result('no')
        return False

def FindProgram(context, var, tool):
    if tool:
        context.Message("Searching {} executable ... ".format(var))

        tool_path = context.env.Which(tool)
        if tool_path:
            context.env[var] = tool_path[0]
        else:
            context.env[var] = tool

        context.Result(context.env[var])

    return True

def FindTool(context, var, toolchains, commands,
             compiler_dir=None, prepend_path=[], required=True):
    env = context.env

    context.Message("Searching {} executable ... ".format(var))

    if env.HasArgument(var):
        context.Result(env[var])
        return True

    toolchains = list(toolchains)

    for tc in list(toolchains):
        if '-pc-' in tc:
            toolchains += [tc.replace('-pc-', '-')]

        if 'android' in tc:
            m = re.search(r'((android(eabi)?)\d+)$', tc)
            if m:
                tc = tc.replace(m.group(1), m.group(2))
                toolchains += [tc]
            toolchains += [tc.replace('armv7a', 'arm')]

    if '' in toolchains:
        toolchains.remove('')
        toolchains.append('')

    if compiler_dir:
        toolchains.append(compiler_dir)

    found = False

    for tool_prefix in toolchains:
        for tool_cmd, tool_ver in commands:
            if isinstance(tool_cmd, list):
                tool_name = tool_cmd[0]
                tool_flags = tool_cmd[1:]
            else:
                tool_name = tool_cmd
                tool_flags = []

            if not tool_prefix:
                tool = tool_name
            elif tool_prefix == compiler_dir:
                tool = os.path.join(compiler_dir, tool_name)
            else:
                tool = '{}-{}'.format(tool_prefix, tool_name)

            if tool_ver:
                search_versions = [
                    tool_ver[:3],
                    tool_ver[:2],
                    tool_ver[:1],
                ]

                default_ver = env.ParseCompilerVersion(tool)

                if default_ver and default_ver[:len(tool_ver)] == tool_ver:
                    search_versions += [default_ver]

                for ver in reversed(sorted(set(search_versions))):
                    versioned_tool = '{}-{}'.format(tool, '.'.join(map(str, ver)))
                    if env.Which(versioned_tool, prepend_path):
                        tool = versioned_tool
                        break

            tool_path = env.Which(tool, prepend_path)
            if not tool_path:
                continue

            if tool_ver:
                actual_ver = env.ParseCompilerVersion(tool_path[0])
                if actual_ver:
                    actual_ver = actual_ver[:len(tool_ver)]

                actual_ver_short = actual_ver if len(actual_ver)<=2 else actual_ver[:2]
                tool_ver_short = tool_ver if len(tool_ver)<=2 else tool_ver[:2]

                if actual_ver_short != tool_ver_short:
                    env.Die(
                        ("problem detecting {}: "
                         "found '{}', which reports version {}, but expected version {}").format(
                            var,
                            tool_path[0],
                            '.'.join(map(str, actual_ver)) if actual_ver else '<unknown>',
                            '.'.join(map(str, tool_ver))))

            env[var] = tool_path[0]
            if tool_flags:
                env[var + 'FLAGS'] = ' '.join(tool_flags)

            found = True
            break

        if found:
            break

    if not found:
        if required:
            env.Die("can't find {} executable".format(var))
        else:
            env[var] = None
            context.Result('not found')
            return False

    message = env[var]
    realpath = os.path.realpath(env[var])
    if realpath != env[var]:
        message += ' ({})'.format(realpath)

    context.Result(message)
    return True

def FindClangFormat(context):
    env = context.env

    context.Message("Searching for clang-format ... ")

    if env.HasArgument('CLANG_FORMAT'):
        context.Result(env['CLANG_FORMAT'])
        return True

    min_ver = 10
    max_ver = 99

    def _checkver(exe):
        ver_str = env.ParseToolVersion('{exe} --version'.format(exe=exe))
        try:
            ver = tuple(map(int, ver_str.split('.')))
            return (min_ver, 0, 0) <= ver <= (max_ver, 99, 99)
        except:
            return False

    clang_format = env.Which('clang-format')
    if clang_format and _checkver(clang_format[0]):
        env['CLANG_FORMAT'] = clang_format[0]
        context.Result(env['CLANG_FORMAT'])
        return True

    for ver in range(min_ver,max_ver+1):
        clang_format = env.Which('clang-format-{}'.format(ver))
        if clang_format and _checkver(clang_format[0]):
            env['CLANG_FORMAT'] = clang_format[0]
            context.Result(env['CLANG_FORMAT'])
            return True

        llvmdir = _get_llvm_dir((ver,))
        if llvmdir:
            clang_format = os.path.join(llvmdir, 'clang-format')
            if _checkver(clang_format):
                env['CLANG_FORMAT'] = clang_format
                context.Result(env['CLANG_FORMAT'])
                return True

    env.Die("can't find clang-format >= {} and <= {}".format(min_ver, max_ver))

def FindLLVMDir(context, version):
    context.Message(
        "Searching PATH for llvm {} ... ".format('.'.join(map(str, version))))

    llvmdir = _get_llvm_dir(version)
    if llvmdir:
        context.env['ENV']['PATH'] += ':' + llvmdir
        context.Result(llvmdir)
        return True

    context.Result('not found')
    return True

def FindLibDir(context, prefix, host):
    context.Message("Searching for system library directory ... ")

    def _libdirs(host):
        dirs = ['lib/' + host]
        if 'x86_64-pc-linux-gnu' == host:
            dirs += ['lib/x86_64-linux-gnu']
        if 'x86_64' in host:
            dirs += ['lib64']
        dirs += ['lib']
        return dirs

    for d in _libdirs(host):
        libdir = os.path.join(prefix, d)
        if os.path.isdir(libdir):
            break

    context.env['ROC_SYSTEM_LIBDIR'] = libdir
    context.Result(libdir)
    return True

def FindConfigGuess(context):
    context.Message('Searching CONFIG_GUESS script ... ')

    if context.env.HasArgument('CONFIG_GUESS'):
        context.Result(context.env['CONFIG_GUESS'])
        return True

    prefixes = [
        '/usr',
        '/usr/local',
        '/usr/local/Cellar',
    ]

    dirs = [
        'share/gnuconfig',
        'share/misc',
        'share/automake-*',
        'automake/*/share/automake-*',
        'share/libtool/build-aux',
        'libtool/*/share/libtool/build-aux',
        'lib/php/build',
        'lib/php/*/build',
    ]

    for p in prefixes:
        for d in dirs:
            for f in context.env.Glob(os.path.join(p, d, 'config.guess')):
                path = str(f)
                if not os.access(path, os.X_OK):
                    continue

                context.env['CONFIG_GUESS'] = path
                context.Result(path)

                return True

    context.Result('not found')
    return False

def FindPkgConfig(context, toolchain):
    env = context.env

    context.Message('Searching PKG_CONFIG ... ')

    if env.HasArgument('PKG_CONFIG'):
        context.Result(env['PKG_CONFIG'])
        return True

    # https://autotools.io/pkgconfig/cross-compiling.html
    # http://tiny.cc/lh6upz
    if toolchain:
        pkg_config_cmd = toolchain + '-pkg-config'
        if env.Which(pkg_config_cmd):
            env['PKG_CONFIG'] = env.Which(pkg_config_cmd)[0]
            context.Result(env['PKG_CONFIG'])
            return True

        if 'PKG_CONFIG_PATH' in os.environ \
            or 'PKG_CONFIG_LIBDIR' in os.environ \
            or 'PKG_CONFIG_SYSROOT_DIR' in os.environ:
            if env.Which('pkg-config'):
                env['PKG_CONFIG'] = env.Which('pkg-config')[0]
                context.Result(env['PKG_CONFIG'])
                return True

        context.Result('not found')
        return False

    if env.Which('pkg-config'):
        env['PKG_CONFIG'] = env.Which('pkg-config')[0]
        context.Result(env['PKG_CONFIG'])
        return True

    context.Result('not found')
    return False

def FindPkgConfigPath(context, prefix):
    if not prefix.endswith(os.sep):
        prefix += os.sep

    env = context.env

    context.Message("Searching PKG_CONFIG_PATH ...")

    if env.HasArgument('PKG_CONFIG_PATH'):
        context.Result(env['PKG_CONFIG_PATH'])
        return True

    pkg_config = env.get('PKG_CONFIG', None)
    if pkg_config:
        pkg_config_path = env.GetCommandOutput(
            '{pkg_config_cmd} --variable pc_path pkg-config'.format(
                pkg_config_cmd=quote(pkg_config)))
        if pkg_config_path:
            env['PKG_CONFIG_PATH'] = pkg_config_path
            context.Result(env['PKG_CONFIG_PATH'])
            return True

    # https://linux.die.net/man/1/pkg-config the default is libdir/pkgconfig
    env['PKG_CONFIG_PATH'] = os.path.join(env['ROC_SYSTEM_LIBDIR'], 'pkgconfig')
    context.Result(env['PKG_CONFIG_PATH'])
    return True

def AddPkgConfigDependency(context, package, flags,
                           add_prefix=None, exclude_from_pc=False, exclude_libs=[]):
    context.Message("Searching pkg-config package {} ...".format(package))

    env = context.env

    pkg_config = env.get('PKG_CONFIG', None)
    if not pkg_config:
        context.Result('pkg-config not available')
        return False

    cmd = []
    pkg_config_path = _compose_pkg_config_path(env, add_prefix)
    if pkg_config_path:
        cmd = ['env', 'PKG_CONFIG_PATH={}'.format(pkg_config_path)]

    cmd += [pkg_config, package, '--silence-errors'] + flags.split()
    try:
        old_libs = env['LIBS'][:]
        old_dirs = env['LIBPATH'][:]

        env.ParseConfig(cmd)

        new_libs = env['LIBS'][:]
        new_dirs = env['LIBPATH'][:]

        for lib in exclude_libs:
            if lib not in old_libs and lib in new_libs:
                env['LIBS'].remove(lib)

        for n, libpath in enumerate(new_dirs):
            if not libpath in old_dirs:
                env['LIBPATH'][n] = _fix_brew_libpath(libpath)

        if not exclude_from_pc:
            if '_DEPS_PCFILES' not in env.Dictionary():
                env['_DEPS_PCFILES'] = []
            env.AppendUnique(_DEPS_PCFILES=[package])
    except:
        context.Result('not found')
        return False

    context.Result('yes')
    return True

def init(env):
    env.CustomTests = {
        'CheckLibWithHeaderExt': CheckLibWithHeaderExt,
        'CheckProg': CheckProg,
        'CheckCanRunProgs': CheckCanRunProgs,
        'CheckCompilerOptionSupported': CheckCompilerOptionSupported,
        'FindProgram': FindProgram,
        'FindTool': FindTool,
        'FindClangFormat': FindClangFormat,
        'FindLLVMDir': FindLLVMDir,
        'FindLibDir': FindLibDir,
        'FindConfigGuess': FindConfigGuess,
        'FindPkgConfig': FindPkgConfig,
        'FindPkgConfigPath': FindPkgConfigPath,
        'AddPkgConfigDependency': AddPkgConfigDependency,
    }



import scons_plugin.arguments
import scons_plugin.commands
import scons_plugin.config
import scons_plugin.distfiles
import scons_plugin.osutils
import scons_plugin.parallel
import scons_plugin.parsers
import scons_plugin.pretty
import scons_plugin.sconsutils
import scons_plugin.systemdeps
import scons_plugin.tests
import scons_plugin.thirdparty
import scons_plugin.wrapper

# workaround for python3
import SCons.Subst
try:
    SCons.Subst.Literal.__hash__ = lambda self: hash(str(self))
except:
    pass

def generate(env):
    modules = [
        scons_plugin.arguments,
        scons_plugin.commands,
        scons_plugin.config,
        scons_plugin.distfiles,
        scons_plugin.osutils,
        scons_plugin.parallel,
        scons_plugin.parsers,
        scons_plugin.pretty,
        scons_plugin.sconsutils,
        scons_plugin.systemdeps,
        scons_plugin.tests,
        scons_plugin.thirdparty,
        scons_plugin.wrapper,
    ]
    for m in modules:
        m.init(env)

def exists(env):
    return 1



import os
import os.path
import fnmatch

try:
    from shlex import quote
except:
    from pipes import quote

def _versioned_thirdparty(env, name, versions):
    if not name in versions:
        env.Die("unknown 3rdparty '{}'", name)
    return '{}-{}'.format(name, versions[name])

def _build_thirdparty(env, versions, name, deps, is_native):
    versioned_name = _versioned_thirdparty(env, name, versions)

    versioned_deps = []
    for dep in deps:
        versioned_deps.append(_versioned_thirdparty(env, dep, versions))

    env_vars = [
        'CXX='    + quote(env['CXX']),
        'CXXLD='  + quote(env['CXXLD']),
        'CC='     + quote(env['CC']),
        'CCLD='   + quote(env['CCLD']),
        'AR='     + quote(env['AR']),
        'RANLIB=' + quote(env['RANLIB']),
    ]

    if 'COMPILER_LAUNCHER' in env.Dictionary():
        env_vars += ['COMPILER_LAUNCHER=' + quote(env['COMPILER_LAUNCHER'])]

    if 'PKG_CONFIG' in env.Dictionary():
        env_vars += ['PKG_CONFIG=' + quote(env['PKG_CONFIG'])]

    project_root = env.Dir('#').srcnode().abspath
    build_root = env.Dir(env['ROC_THIRDPARTY_BUILDDIR']).abspath
    thirdparty_dir = os.path.join(build_root, versioned_name)
    distfiles_dir = os.path.join('3rdparty', '_distfiles')

    if not os.path.exists(os.path.join(thirdparty_dir, 'commit')):
        saved_cwd = os.getcwd()
        os.chdir(project_root)

        cmd = [
            quote(env.GetPythonExecutable()), 'scripts/scons_helpers/build-3rdparty.py',
            '--root-dir', quote(os.path.abspath(project_root)),
            '--work-dir', quote(os.path.relpath(build_root, project_root)),
            '--dist-dir', quote(os.path.relpath(distfiles_dir, project_root)),
            ]

        if not is_native:
            cmd += [
                '--build',     quote(env['ROC_BUILD']),
                '--host',      quote(env['ROC_HOST']),
                '--toolchain', quote(env['ROC_TOOLCHAIN']),
            ]
        else:
            cmd += [
                '--build', quote(env['ROC_BUILD']),
                '--host',  quote(env['ROC_BUILD']),
            ]

        cmd += [
            '--variant', quote(env['ROC_THIRDPARTY_VARIANT']),
            '--package', quote(versioned_name),
        ]

        if versioned_deps:
            cmd += ['--deps', ' '.join(versioned_deps)]

        if env_vars:
            cmd += ['--vars', ' '.join(env_vars)]

        if env['ROC_PLATFORM'] == 'android' and env['ROC_ANDROID_PLATFORM']:
            cmd += ['--android-platform', env['ROC_ANDROID_PLATFORM']]

        if env['ROC_PLATFORM'] == 'darwin' and env['ROC_MACOS_PLATFORM']:
            cmd += ['--macos-platform', env['ROC_MACOS_PLATFORM']]

        if env['ROC_PLATFORM'] == 'darwin' and env['ROC_MACOS_ARCH']:
            cmd += ['--macos-arch', ' '.join(env['ROC_MACOS_ARCH'])]

        if env.Execute(
            ' '.join(cmd),
            cmdstr = env.PrettyCommand(
                'BUILD', os.path.relpath(thirdparty_dir, project_root), 'yellow')):

            log_file = os.path.join(thirdparty_dir, 'build.log')
            message = "can't make '{}', see '{}' for details".format(
                versioned_name, os.path.relpath(log_file, project_root))

            if os.environ.get('CI', '') in ['1', 'true']:
                try:
                    with open(log_file) as fp:
                        message += "\n\n" + fp.read()
                except:
                    pass

            env.Die('{}', message)

        os.chdir(saved_cwd)

def _import_thridparty(env, versions, name, includes, libs):
    def _needlib(lib):
        for name in libs:
            if fnmatch.fnmatch(os.path.basename(lib), 'lib{}.*'.format(name)):
                return True
        return False

    versioned_name = _versioned_thirdparty(env, name, versions)

    build_root = env.Dir(env['ROC_THIRDPARTY_BUILDDIR']).abspath

    if not includes:
        includes = ['']

    for s in includes:
        incdir = '{}/{}/include'.format(build_root, versioned_name)
        if s:
            incdir += '/' + s

        if os.path.isdir(env.Dir(incdir).abspath):
            env.Prepend(CPPPATH=[incdir])

    libdir = '{}/{}/lib'.format(build_root, versioned_name)
    rpathdir = '{}/{}/rpath'.format(build_root, versioned_name)

    if os.path.isdir(env.Dir(libdir).abspath):
        env.Prepend(LIBPATH=[libdir])

        if env['ROC_PLATFORM'] == 'linux':
            env.Prepend(LINKFLAGS=[
                '-Wl,-rpath-link,{}'.format(env.Dir(rpathdir).path),
            ])

        for lib in env.GlobRecursive(env.Dir(libdir).abspath, 'lib*'):
            if _needlib(lib.path):
                env.Prepend(LIBS=[env.File(lib)])
                env.Append(_THIRDPARTY_LIBS=[lib.abspath])

def ParseThirdPartyList(env, s):
    ret = dict()
    if s:
        for t in s.split(','):
            tokens = t.split(':', 1)

            name = tokens[0]
            ver = None
            if name != 'all' and len(tokens) == 2:
                ver = tokens[1]

            ret[name] = ver
    return ret.items()

def BuildThirdParty(env, versions, name, deps=[], includes=[], libs=['*'], is_native=False):
    _build_thirdparty(
        env, versions, name, deps, is_native)

    _import_thridparty(
        env, versions, name, includes, libs)

def GetThirdPartyExecutable(env, versions, name, exe_name):
    return env.File(
        '{}/{}/bin/{}{}'.format(
            env['ROC_THIRDPARTY_BUILDDIR'],
            _versioned_thirdparty(env, name, versions),
            exe_name,
            env['PROGSUFFIX']))

def GetThirdPartyStaticLibs(env):
    all_libs = env.get('_THIRDPARTY_LIBS', [])
    static_libs = []

    for lib in all_libs:
        if lib.endswith(env['LIBSUFFIX']):
            static_libs.append(lib)

    return static_libs

def init(env):
    env.AddMethod(ParseThirdPartyList, 'ParseThirdPartyList')
    env.AddMethod(BuildThirdParty, 'BuildThirdParty')
    env.AddMethod(GetThirdPartyExecutable, 'GetThirdPartyExecutable')
    env.AddMethod(GetThirdPartyStaticLibs, 'GetThirdPartyStaticLibs')



import os
import textwrap

# Python 2 compatibility.
try:
    from shlex import quote
except:
    from pipes import quote

# Generate .pc file for roc.
# This function uses information accumulated by:
#  - AddPkgConfigDependency - list of .pc files of system dependencies
#  - AddManualDependency - list of libs and search paths of system dependencies
#    (for cases when system dependency does not have .pc file)
# All this information is included into resulting .pc file.
# To work properly, GeneratePkgConfig should be called after aforementioned functions.
def GeneratePkgConfig(env, build_dir, filename, prefix, libdir, name, desc, url, version):
    target = os.path.join(build_dir, filename)

    dep_list = env.get('_DEPS_PCFILES', [])

    lib_list = \
      ['roc'] + \
      env.get('_DEPS_LIBS', [])

    libdir_list = \
      [libdir] + \
      env.get('_DEPS_LIBPATH', [])

    incdir_list = \
      [prefix+'/include'] + \
      env.get('_DEPS_CPPPATH', [])

    src = textwrap.dedent("""\
        prefix={prefix}
        exec_prefix=${{prefix}}

        Name: {name}
        Description: {desc}
        URL: {url}
        Version: {version}
        Requires: {deps}

        Libs: {libdirs} {libs}
        Cflags: {incdirs}
    """).format(
        prefix=prefix,
        name=name,
        desc=desc,
        url=url,
        version=version,
        deps=' '.join(dep_list),
        libs=' '.join(['-l'+s for s in lib_list]),
        libdirs=' '.join(['-L'+quote(s) for s in libdir_list]),
        incdirs=' '.join(['-I'+quote(s) for s in incdir_list]))

    def write_file(target, source, env):
        f = open(target[0].path, 'w')
        f.write(src)
        f.close()

    env.Command(target, [], [
        env.Action(write_file, env.PrettyCommand('GEN', env.File(target).path, 'purple')),
        ])

    return env.File(target)

# Add system dependency manually. Used as fallback when AddPkgConfigDependency did not
# work or can't be used.
# If libs arg is provided, libs are added to linker options.
# If prefix arg is provided, and it defines a dir which has include and lib subdirs,
# those subdirs are added to the headers and library search paths.
# This function also automatically informs GeneratePkgConfig that correspondings
# paths and libs should be added to .pc file.
def AddManualDependency(env, libs=[], prefix=None, exclude_from_pc=False):
    def _append_var(var, values):
        if var not in env.Dictionary():
            env[var] = []
        env.AppendUnique(**{var: values})

    if libs:
        _append_var('LIBS', libs)
        if not exclude_from_pc:
            _append_var('_DEPS_LIBS', libs)

    if prefix:
        lib_dir = os.path.join(prefix, 'lib')
        if os.path.isdir(lib_dir):
            _append_var('LIBPATH', [lib_dir])
            if not exclude_from_pc:
                _append_var('_DEPS_LIBPATH', [lib_dir])

        inc_dir = os.path.join(prefix, 'include')
        if os.path.isdir(inc_dir):
            _append_var('CPPPATH', [inc_dir])
            if not exclude_from_pc:
                _append_var('_DEPS_CPPPATH', [inc_dir])

# Find and return path to the package (its install prefix directory) of the latest version
# of `pkg_name` package installed by brew. If not found, None is returned.
def FindBrewPackage(env, pkg_name):
    if not env.Which('brew'):
        return None

    brew_prefix = env.GetCommandOutput('brew --prefix')
    if not brew_prefix:
        return None

    # if pkg-config is not from brew, then don't try to use brew packages implicitly,
    # that's likely not what the user wants
    pkg_config = env.get('PKG_CONFIG', None)
    if not pkg_config or \
       not env.Which(pkg_config) or \
       not os.path.abspath(env.Which(pkg_config)[0]).startswith(
           os.path.abspath(brew_prefix)+os.sep):
        return None

    # $(brew --prefix)/opt/some_pkg should be a link to the latest version of some_pkg,
    # e.g. ../Cellar/some_pkg@1.2/1.2.3/
    pkg_prefix = os.path.join(brew_prefix, 'opt', pkg_name)
    if not os.path.isdir(pkg_prefix):
        return None

    return pkg_prefix

def init(env):
    env.AddMethod(GeneratePkgConfig, 'GeneratePkgConfig')
    env.AddMethod(AddManualDependency, 'AddManualDependency')
    env.AddMethod(FindBrewPackage, 'FindBrewPackage')



#! /usr/bin/env python3
import os
import os.path
import subprocess

def format_name(encoding, endian):
    return 'PcmFormat_' + \
        encoding['pcm_encoding'] + '_' + endian['pcm_endian']

def format_array(array, maxlen=8, indent=1):
    if len(array) > maxlen:
        ret = "{\n"
        while array:
            ret += " " * (indent * 2 + 1)
            for n in range(maxlen):
                if array:
                    ret += " %s," % array.pop(0)
            ret += "\n"
        ret += " " * (indent * 2)
        ret += "}"
    else:
        ret = "{ "
        ret += ", ".join(map(str, array))
        ret += " }"

    return ret

def format_blob(blob):
    return format_array(["0x%02x" % b for b in blob], maxlen=12)

def format_samples(samples):
    return format_array(["%+.16f" % s for s in samples], maxlen=4)

def read_samples(datfile):
    samples = []

    with open(datfile) as fp:
        for line in fp.read().splitlines():
            line = line.strip()
            if line.startswith(';'):
                continue
            time, value = map(float, line.split())
            samples.append(value)

    return samples

def synthesize_samples(datfile):
    cmd = ' '.join(map(str, [
        'sox', '-V0', '-R',
        '--null',
        '--rate', 48000,
        '--channels', 1,
        datfile,
        'synth', '0.005', 'sin', '1000',
        ]))
    subprocess.call(cmd, shell=True)

def generate_blob(encoding, endian, datfile, rawfile):
    cmd = ' '.join(map(str, [
        'sox', '-V0', '-R',
        datfile,
        '--type', 'raw',
        '--encoding', encoding['sox_encoding'],
        '--bits', encoding['depth'],
        '--endian', endian['sox_endian'],
        rawfile,
        ]))
    subprocess.call(cmd, shell=True)

def generate_header(encoding, endian, name, datfile, rawfile, hdrfile):
    with open(rawfile, 'rb') as fp:
        rawdata = fp.read()

    samples = read_samples(datfile)

    guard = 'ROC_AUDIO_TEST_SAMPLES_' + hdrfile.upper().replace('.', '_') + '_'

    with open(hdrfile, 'w') as fp:
        fp.write(f'''
/*
 * THIS FILE IS AUTO-GENERATED USING `generate_samples.py'
 */

#ifndef {guard}
#define {guard}

#include "test_samples/sample_info.h"

namespace roc {{
namespace audio {{
namespace test {{

static SampleInfo sample_{name} = {{
  /* name */ "{name}",

  /* format */ {format_name(encoding, endian)},

  /* num_samples */ {len(samples)},
  /* samples     */ {format_samples(samples)},

  /* num_bytes */ {len(rawdata)},
  /* bytes     */ {format_blob(rawdata)}
}};

}} // namespace test
}} // namespace audio
}} // namespace roc

#endif // {guard}
'''.lstrip())

encodings = [
    {
        'pcm_encoding': 'SInt8',
        'sox_encoding': 'signed-integer',
        'depth': 8,
    },
    {
        'pcm_encoding': 'UInt8',
        'sox_encoding': 'unsigned-integer',
        'depth': 8,
    },
    {
        'pcm_encoding': 'SInt16',
        'sox_encoding': 'signed-integer',
        'depth': 16,
    },
    {
        'pcm_encoding': 'UInt16',
        'sox_encoding': 'unsigned-integer',
        'depth': 16,
    },
    {
        'pcm_encoding': 'SInt24',
        'sox_encoding': 'signed-integer',
        'depth': 24,
    },
    {
        'pcm_encoding': 'UInt24',
        'sox_encoding': 'unsigned-integer',
        'depth': 24,
    },
    {
        'pcm_encoding': 'SInt32',
        'sox_encoding': 'signed-integer',
        'depth': 32,
    },
    {
        'pcm_encoding': 'UInt32',
        'sox_encoding': 'unsigned-integer',
        'depth': 32,
    },
    {
        'pcm_encoding': 'Float32',
        'sox_encoding': 'floating-point',
        'depth': 32,
    },
]

endians = [
    {
        'pcm_endian': 'Be',
        'sox_endian': 'big',
    },
    {
        'pcm_endian': 'Le',
        'sox_endian': 'little',
    },
]

os.chdir(os.path.dirname(os.path.abspath(__file__)))

datfile = 'original_samples.dat'

print(f'generating {datfile}')
synthesize_samples(datfile)

for encoding in encodings:
    for endian in endians:
        output = '_'.join([
            'pcm',
            encoding['pcm_encoding'].lower(),
            endian['pcm_endian'][0].lower() + 'e',
            ])

        print(f'generating {output}')

        rawfile = f'{output}.blob'
        hdrfile = f'{output}.h'

        generate_blob(encoding, endian, datfile, rawfile)
        generate_header(encoding, endian, output, datfile, rawfile, hdrfile)



from __future__ import print_function

import os
import os.path
import re

os.chdir(os.path.dirname(os.path.realpath(__file__)))

fp = open('../../src/public_api/include/roc/version.h')

data = fp.read()

m = re.search(r"""^#define ROC_VERSION_MAJOR (\d+)$""", data, re.MULTILINE)
major = m.group(1)

m = re.search(r"""^#define ROC_VERSION_MINOR (\d+)$""", data, re.MULTILINE)
minor = m.group(1)

m = re.search(r"""^#define ROC_VERSION_PATCH (\d+)$""", data, re.MULTILINE)
patch = m.group(1)

print('.'.join([major, minor, patch]))



import fnmatch
import os
import os.path
import shutil
import subprocess
import sys

def GetPythonExecutable(env):
    base = os.path.basename(sys.executable)
    path = env.Which(base)
    if path and path[0] == sys.executable:
        return base
    else:
        return sys.executable

def GetCommandOutput(env, command):
    try:
        with open(os.devnull, 'w') as null:
            proc = subprocess.Popen(command,
                                    shell=True,
                                    stdin=null,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT,
                                    env=env['ENV'])
            lines = [s.decode() for s in proc.stdout.readlines()]
            output = str(' '.join(lines).strip())
            proc.terminate()
            return output
    except:
        return None

def GlobRecursive(env, dirs, patterns, exclude=[]):
    if not isinstance(dirs, list):
        dirs = [dirs]

    if not isinstance(patterns, list):
        patterns = [patterns]

    if not isinstance(exclude, list):
        exclude = [exclude]

    matches = []

    for pattern in patterns:
        for root in dirs:
            if type(root) is not str or not os.path.isabs(root):
                root = env.Dir(root).srcnode().abspath
            for root, dirnames, filenames in os.walk(root):
                for names in [dirnames, filenames]:
                    for name in fnmatch.filter(names, pattern):
                        cwd = env.Dir('.').srcnode().abspath

                        abspath = os.path.join(root, name)
                        relpath = os.path.relpath(abspath, cwd)

                        for ex in exclude:
                            if fnmatch.fnmatch(relpath, ex):
                                break
                            if fnmatch.fnmatch(os.path.basename(relpath), ex):
                                break
                        else:
                            if names is dirnames:
                                matches.append(env.Dir(relpath))
                            else:
                                matches.append(env.File(relpath))

    return matches

def GlobFiles(env, pattern):
    ret = []
    for path in env.Glob(pattern):
        if os.path.isfile(path.srcnode().abspath):
            ret.append(path)
    return ret

def GlobDirs(env, pattern):
    ret = []
    for path in env.Glob(pattern):
        if os.path.isdir(path.srcnode().abspath):
            ret.append(path)
    return ret

def Which(env, prog, prepend_path=[]):
    def _getenv(env, name, default):
        if name in env['ENV']:
            return env['ENV'][name]
        return os.environ.get(name, default)

    def _which(env, prog, mode, searchpath):
        result = []
        exts = list(filter(None, _getenv(env, 'PATHEXT', '').split(os.pathsep)))
        for p in searchpath.split(os.pathsep):
            p = os.path.join(p, prog)
            if os.access(p, mode):
                result.append(p)
            for e in exts:
                pext = p + e
                if os.access(pext, mode):
                    result.append(pext)
        return result

    if os.access(prog, os.X_OK):
        return [prog]

    searchpath = _getenv(env, 'PATH', os.defpath)
    if prepend_path:
        searchpath = os.pathsep.join(prepend_path) + os.pathsep + searchpath

    paths = []
    try:
        path = shutil.which(prog, os.X_OK, searchpath)
        if path:
            paths = [path]
    except:
        paths = _which(env, prog, os.X_OK, searchpath)

    return paths

def init(env):
    env.AddMethod(GetPythonExecutable, 'GetPythonExecutable')
    env.AddMethod(GetCommandOutput, 'GetCommandOutput')
    env.AddMethod(GlobRecursive, 'GlobRecursive')
    env.AddMethod(GlobFiles, 'GlobFiles')
    env.AddMethod(GlobDirs, 'GlobDirs')
    env.AddMethod(Which, 'Which')



#!/usr/bin/env python3
#
# usage:
# roc-test-audio -v -g resampler -n upscale_downscale_mono |& \
#  ./scripts/plot_resampler_test_dump.py
#
import pylab
import numpy
import re
import fileinput

orig = []
proc = []

for line in fileinput.input():
    m = re.search('roc_test: dump (\S+) (\S+)', line)
    if m:
        orig.append(float(m.group(1)))
        proc.append(float(m.group(2)))

orig = numpy.array(orig)
proc = numpy.array(proc)

_, px = pylab.subplots(2, 1, sharex=True)

px[0].plot(orig, '.-', label='orig')
px[0].plot(proc, '.-', label='proc')
px[0].legend()
px[0].grid()

px[1].plot(numpy.abs(orig-proc), '.-', label='abs(diff)')
px[1].legend()
px[1].grid()

pylab.show()



from __future__ import print_function

import sys
import os
import threading
import subprocess

try:
    timeout = int(sys.argv[1])
    cmd = sys.argv[2:]
    if len(cmd) == 0:
        raise ValueError()
except:
   print("usage: timeout-run.py TIMEOUT COMMAND...", file=sys.stderr)
   exit(1)

proc = subprocess.Popen(cmd)

def trap():
    print("timeout of {}s expired, aborting".format(timeout), file=sys.stderr)
    try:
        proc.terminate()
    except:
        pass

timer = threading.Timer(timeout, trap)
timer.start()

try:
    ret = proc.wait()
    if ret >= 0:
        os._exit(ret)
    else:
        os.kill(os.getpid(), -ret)
except:
    proc.terminate()
    os._exit(127)



import datetime
import fnmatch
import os
import re
import shutil
import sys
import tempfile
import textwrap

copyright_str = textwrap.dedent('''
    /*
     * Copyright (c) {} Roc Streaming authors
     *
     * This Source Code Form is subject to the terms of the Mozilla Public
     * License, v. 2.0. If a copy of the MPL was not distributed with this
     * file, You can obtain one at http://mozilla.org/MPL/2.0/.
     */
''').format(datetime.datetime.now().year)

def is_header(path):
    return re.search(r'\.h$', path)

def is_test(path):
    _, basename = os.path.split(path)
    for s in [basename, os.path.basename(os.path.dirname(path))]:
        if s.startswith('test_'):
            return True
    return False

def is_public_api(path):
    rootname = os.path.basename(
        os.path.dirname(
            os.path.dirname(
                os.path.abspath(path))))
    return rootname == 'public_api'

def make_guard(path):
    dirpath, basename = os.path.split(path)
    dirname = os.path.basename(dirpath)
    if is_public_api(path):
        if is_test(path):
            arr = ['roc', 'public_api', dirname, basename]
        else:
            arr = ['roc', 'public_api', basename]
    else:
        arr = [dirname, basename]
    while not arr[0].startswith('roc_') and arr[0] != 'roc':
        dirpath = os.path.dirname(dirpath)
        arr = [os.path.basename(dirpath)] + arr
    return '_'.join(arr).replace('.', '_').upper() + '_'

def make_doxygen_path(path):
    path = '/'.join(path.split(os.sep))
    path = re.sub(r'^\.?/', '', path)
    return '@file ' + path

def make_doxygen_brief(text):
    if not text.endswith('.'):
        text += '.'
    return '@brief ' + text

def format_file(output, path):
    def fprint(s):
        output.write(s + '\n')

    with open(path) as fp:
        lines = fp.read().splitlines()

    has_copyright, has_doxygen, has_guard, is_autogen = False, False, False, False

    section = 'copyright'
    brief = 'TODO'

    original_lines = []
    body_start = True

    if is_header(path):
        while re.match(r'^\s*$', lines[-1]):
            lines.pop()

        if re.match(r'^\s*#\s*endif.*$', lines[-1]):
            lines.pop()

        while re.match(r'^\s*$', lines[-1]):
            lines.pop()

        for line in lines:
            m = re.search(r'@brief\s+(.*)', line)
            if m:
                brief = m.group(1)
                break

    while lines:
        line = lines.pop(0)

        if section in ['copyright', 'doxygen', 'guard']:
            if re.match(r'^\s*$', line):
                continue

        if section == 'copyright':
            original_lines += [line]

            if re.match(r'^\s*/?\*', line):
                if re.match(r'^\s*/?\*.*AUTO-GENERATED.*', line):
                    is_autogen = True

                if re.match(r'^\s*/?\*\s*(Copyright|Mozilla)', line):
                    has_copyright = True

                if re.match(r'^\s*\*/', line):
                    if is_autogen or has_copyright:
                        for p in original_lines:
                            fprint(p)
                        fprint('')
                    else:
                        fprint(copyright_str.strip())
                        fprint('')

                        if not has_copyright:
                            for p in original_lines:
                                fprint(p)
                        fprint('')

                    section = 'doxygen' if is_header(path) else 'body'
                    continue
            else:
                fprint(copyright_str.strip())
                fprint('')

                lines = original_lines + lines
                section = 'doxygen' if is_header(path) else 'body'

                continue

        if section == 'doxygen':
            if re.match(r'^\s*/?\*', line) or re.match(r'^\s*//', line):
                continue

            if re.match(r'^\s*//!', line):
                if not is_header(path) or is_test(path) or is_autogen:
                    continue

                if re.match(r'^\s*//!\s*@file', line):
                    fprint('//! {}'.format(make_doxygen_path(path)))
                else:
                    fprint(line)

                has_doxygen = True
                continue
            else:
                if is_public_api(path) or is_test(path) or is_autogen:
                    section = 'guard' if is_header(path) else 'body'
                else:
                    if not has_doxygen:
                        if is_header(path):
                            fprint('//! {}'.format(make_doxygen_path(path)))
                            fprint('//! {}'.format(make_doxygen_brief(brief)))
                            section = 'guard'
                        else:
                            section = 'body'

                    if is_header(path):
                        fprint('')

        if section == 'guard':
            m = re.match(r'#\s*(ifndef|define)', line)
            if m:
                has_guard = True
                fprint('#{} {}'.format(m.group(1), make_guard(path)))
                continue
            else:
                if not has_guard:
                    fprint('#ifndef {}'.format(make_guard(path)))
                    fprint('#define {}'.format(make_guard(path)))
                fprint('')
                section = 'body'

        if section == 'body':
            if body_start and re.match(r'^\s*$', line):
                continue
            body_start = False
            fprint(line)

    if is_header(path):
        fprint('')
        fprint('#endif // {}'.format(make_guard(path)))

def walk_dir(directory, patterns):
    for root, dirs, files in os.walk(directory):
        for basename in files:
            for pattern in patterns:
                if fnmatch.fnmatch(basename, pattern):
                    filename = os.path.join(root, basename)
                    yield filename
                    break

if len(sys.argv) > 1:
    os.chdir(sys.argv[1])

for path in walk_dir('.', ['*.h', '*.cpp']):
    with tempfile.NamedTemporaryFile('w+') as fp:
        format_file(fp, path)
        fp.flush()
        shutil.copy(fp.name, path)



#! /usr/bin/env python3

# This script generates downmixing tables for ChannelMapperMatrix.
# Its logic is complicated, but the result (C++ file with tables)
# is quite straightforward to read and verify.

# This approach makes C++ code simpler - it just needs to build
# a matrix from tables and apply it to stream.

from collections import OrderedDict
from fnmatch import fnmatch
from itertools import islice
from math import *
import builtins
import os
import os.path
import re
import sys

try:
    import jinja2
except ImportError:
    print('''
error: can't import python module "jinja2", install it using "pip3 install Jinja2"
'''.strip(), file=sys.stderr)
    exit(1)

CHANS = OrderedDict([
    ('FL', 'FrontLeft'),
    ('FLC', 'FrontLeftOfCenter'),
    ('FC', 'FrontCenter'),
    ('FRC', 'FrontRightOfCenter'),
    ('FR', 'FrontRight'),
    ('SL', 'SideLeft'),
    ('SR', 'SideRight'),
    ('BL', 'BackLeft'),
    ('BC', 'BackCenter'),
    ('BR', 'BackRight'),
    ('TFL', 'TopFrontLeft'),
    ('TFR', 'TopFrontRight'),
    ('TML', 'TopMidLeft'),
    ('TMR', 'TopMidRight'),
    ('TBL', 'TopBackLeft'),
    ('TBR', 'TopBackRight'),
    ('LFE', 'LowFrequency'),
])

ORDERS = OrderedDict([
    ('none', {
        'id':        'None',
        'text_name': 'none',
        'chans':     [],
        }),
    ('smpte', {
        'id':        'Smpte',
        'text_name': 'smpte',
        'chans':     'FL FR FC LFE BL BR FLC FRC BC SL SR TFL TFR TML TMR TBL TBR'.split(),
        }),
    ('alsa', {
        'id':        'Alsa',
        'text_name': 'alsa',
        'chans':     'FL FR BL BR FC LFE SL SR BC'.split(),
        }),
])

MASKS = OrderedDict([
    # 1.x
    ('1.0', {
        'id':        'Mono',
        'text_name': 'mono',
        'chans':     'FC'.split(),
        'map_to':    [],
     }),
    ('1.1', {
        'id':        '1_1',
        'chans':     'FC LFE'.split(),
        'map_to':    ['1.1*'],
     }),
    ('1.1-3c', {
        'id':        '1_1_3c',
        'chans':     'FLC FC FRC LFE'.split(),
        'map_to':    ['1.1*'],
     }),
    # 2.x
    ('2.0', {
        'id':        'Stereo',
        'text_name': 'stereo',
        'chans':     'FL FR'.split(),
        'map_to':    [],
     }),
    ('2.1', {
        'id':        '2_1',
        'text_name': 'surround2.1',
        'chans':     'FL FR LFE'.split(),
        'map_to':    ['1.1*', '2.1*'],
     }),
    # 3.x
    ('3.0', {
        'id':        '3_0',
        'text_name': 'surround3.0',
        'chans':     'FL FC FR'.split(),
        'map_to':    [],
     }),
    ('3.1', {
        'id':        '3_1',
        'text_name': 'surround3.1',
        'chans':     'FL FC FR LFE'.split(),
        'map_to':    ['1.1*', '2.1*', '3.1*'],
     }),
    ('3.1-3c', {
        'id':        '3_1_3c',
        'chans':     'FL FLC FC FRC FR LFE'.split(),
        'map_to':    ['1.1*', '2.1*', '3.1*'],
     }),
    # 4.x
    ('4.0', {
        'id':        '4_0',
        'text_name': 'surround4.0',
        'chans':     'FL FR BL BR'.split(),
        'map_to':    [],
     }),
    ('4.1', {
        'id':        '4_1',
        'text_name': 'surround4.1',
        'chans':     'FL FR BL BR LFE'.split(),
        'map_to':    ['2.1*', '3.1*', '4.1*'],
     }),
    # 5.x
    ('5.0', {
        'id':        '5_0',
        'text_name': 'surround5.0',
        'chans':     'FL FC FR BL BR'.split(),
        'map_to':    [],
     }),
    ('5.1', {
        'id':        '5_1',
        'text_name': 'surround5.1',
        'chans':     'FL FC FR BL BR LFE'.split(),
        'map_to':    ['3.1*', '4.1*', '5.1*'],
     }),
    ('5.1-3c', {
        'id':        '5_1_3c',
        'chans':     'FL FLC FC FRC FR BL BR LFE'.split(),
        'map_to':    ['3.1*', '4.1*', '5.1*'],
     }),
    ('5.1.2', {
        'id':        '5_1_2',
        'text_name': 'surround5.1.2',
        'chans':     'FL FC FR BL BR TML TMR LFE'.split(),
        'map_to':    ['3.1*', '4.1*', '5.1*'],
     }),
    ('5.1.2-3c', {
        'id':        '5_1_2_3c',
        'chans':     'FL FLC FC FRC FR BL BR TML TMR LFE'.split(),
        'map_to':    ['3.1*', '4.1*', '5.1*'],
     }),
    ('5.1.4', {
        'id':        '5_1_4',
        'text_name': 'surround5.1.4',
        'chans':     'FL FC FR BL BR TFL TFR TBL TBR LFE'.split(),
        'map_to':    ['3.1*', '4.1*', '5.1*'],
     }),
    ('5.1.4-3c', {
        'id':        '5_1_4_3c',
        'chans':     'FL FLC FC FRC FR BL BR TFL TFR TBL TBR LFE'.split(),
        'map_to':    ['3.1*', '4.1*', '5.1*'],
     }),
    # 6.x
    ('6.0', {
        'id':        '6_0',
        'text_name': 'surround6.0',
        'chans':     'FL FC FR BL BC BR'.split(),
        'map_to':    [],
     }),
    ('6.1', {
        'id':        '6_1',
        'text_name': 'surround6.1',
        'chans':     'FL FC FR BL BC BR LFE'.split(),
        'map_to':    ['4.1*', '5.1*', '6.1*'],
     }),
    ('6.1-3c', {
        'id':        '6_1_3c',
        'chans':     'FL FLC FC FRC FR BL BC BR LFE'.split(),
        'map_to':    ['4.1*', '5.1*', '6.1*'],
     }),
    # 7.x
    ('7.0', {
        'id':        '7_0',
        'text_name': 'surround7.0',
        'chans':     'FL FC FR SL SR BL BR'.split(),
        'map_to':    [],
     }),
    ('7.1', {
        'id':        '7_1',
        'text_name': 'surround7.1',
        'chans':     'FL FC FR SL SR BL BR LFE'.split(),
        'map_to':    ['5.1*', '6.1*', '7.1*'],
     }),
    ('7.1-3c', {
        'id':        '7_1_3c',
        'chans':     'FL FLC FC FRC FR SL SR BL BR LFE'.split(),
        'map_to':    ['5.1*', '6.1*', '7.1*'],
     }),
    ('7.1.2', {
        'id':        '7_1_2',
        'text_name': 'surround7.1.2',
        'chans':     'FL FC FR SL SR BL BR TML TMR LFE'.split(),
        'map_to':    ['5.1*', '6.1*', '7.1*'],
     }),
    ('7.1.2-3c', {
        'id':        '7_1_2_3c',
        'chans':     'FL FLC FC FRC FR SL SR BL BR TML TMR LFE'.split(),
        'map_to':    ['5.1*', '6.1*', '7.1*'],
     }),
    ('7.1.4', {
        'id':        '7_1_4',
        'text_name': 'surround7.1.4',
        'chans':     'FL FC FR SL SR BL BR TFL TFR TBL TBR LFE'.split(),
        'map_to':    ['5.1*', '6.1*', '7.1*'],
     }),
    ('7.1.4-3c', {
        'id':        '7_1_4_3c',
        'chans':     'FL FLC FC FRC FR SL SR BL BR TFL TFR TBL TBR LFE'.split(),
        'map_to':    ['5.1*', '6.1*', '7.1*'],
     }),
])

# List all channel names.
def chan_pos_names():
    ret = []
    for ch_name, ch_id in CHANS.items():
        ret += [(ch_id, ch_name)]
    return ret

# List all channel masks that have text names.
def chan_mask_names():
    ret = []
    for mask in MASKS.values():
        if mask.get('text_name', None):
            ret += [(mask['id'], mask['text_name'])]
    return ret

# List all channel orders.
def chan_orders():
    ret = []
    for order in ORDERS.values():
        ret += [(order['id'], order['text_name'], order['chans'])]
    return ret

# List all channel mask pairs for which we should generate mappings.
def chan_mappings():
    def want_mapping(src_name, src_mask, dst_name, dst_mask):
        # If map_to patterns of src don't cover dst, skip.
        # This limits how much lower levels to cover, e.g. there are direct
        # mappings of 7.x to 6.x and 5.x, but not 3.x.
        for pat in src_mask['map_to']:
            if fnmatch(dst_name, pat):
                break
        else:
            return False

        if '3c' not in src_name and '3c' in dst_name:
            # Non-3c is mapped only to non-3c.
            # (e.g. 5.1.2 is mapped to 3.1.2, but not 3.1.2-3c)
            return False

        if '3c' in src_name and '3c' not in dst_name and not src_name.startswith(dst_name):
            # 3c is mapped only to:
            #   - any 3c
            #   - non-3c of same level
            # (e.g. 5.1.2-3c is mapped to 3.1.2-3c and 5.1.2, but not 3.1.2)
            return False

        return True

    ret = []
    for src_idx, (src_name, src_mask) in islice(enumerate(MASKS.items()), 1, None):
        src_start = True
        for dst_idx, (dst_name, dst_mask) in islice(enumerate(MASKS.items()), 0, src_idx):
            # Here we skip some of the redundant mappings.
            # Theoretically we could generate mappings for all mask combinations,
            # but there would be just too much tables.
            # When there is no table for direct conversion, channel mapper will
            # use cascade downmixing/upmixing and combine multiple tables.
            if not want_mapping(src_name, src_mask, dst_name, dst_mask):
                continue

            ret += [(src_start, src_name, src_mask, dst_name, dst_mask)]
            src_start = False

    return ret

# For given dst chan, returns list of src chans with coefficients.
# E.g. for 'FC' it can return [('FL', 1.000), ('FR', 1.000)], depending
# on src mask and dst mask.
#
# This function was designed to produce downmixing tables equivalent to those
# defined in ITU and AC-3 standards, referenced in channel_tables.h.
# It, presumably, replicates heuristics used to define the tables,
# and allows to extend their usage to combinations not covered in documents.
#
# I don't believe that all these formulas are correct, though generated tables
# look sensible. Also, cascade down/up-mixing seems to give the same result
# as the direct mapping, as expected. But this definitely needs more eyeballs.
def map_chan(dst_chan, dst_mask, src_mask):
    def is_f(chan):
        return chan.startswith('F')
    def is_s(chan):
        return chan.startswith('S')
    def is_b(chan):
        return chan.startswith('B')
    def is_t(chan):
        return chan.startswith('T')
    def is_l(chan):
        return chan.endswith('L') or chan.endswith('LC')
    def is_r(chan):
        return chan.endswith('R') or chan.endswith('RC')

    def to_l(chan):
        chan = re.sub('R$', 'L', chan)
        chan = re.sub('RC$', 'LC', chan)
        return chan
    def to_r(chan):
        chan = re.sub('L$', 'R', chan)
        chan = re.sub('LC$', 'RC', chan)
        return chan

    def r_from_l():
        for chan, coeff in map_chan(to_l(dst_chan), dst_mask, src_mask):
            yield (to_r(chan), coeff)

    def in_src(ch):
        return ch in src_mask
    def in_dst(ch):
        return ch in dst_mask

    _1_000 = 1
    _0_707 = 1 / sqrt(2)
    _0_500 = 1 / 2
    _0_354 = 1 / (2 * sqrt(2))

    match dst_chan:
        case 'FL':
            # mix front
            if in_src('FL'):
                yield ('FL', _1_000)
            if in_src('FLC') and not in_dst('FLC'):
                yield ('FLC', _0_707)
            if in_src('FC') and not in_dst('FC'):
                yield ('FC', _0_707)
            # mix surround
            if not in_dst('SL') and not in_dst('BL'):
                # SL and BL
                if in_src('SL') and in_src('BL'):
                    yield ('SL', _0_500)
                    yield ('BL', _0_500)
                elif in_src('SL'):
                    yield ('SL', _0_707)
                elif in_src('BL'):
                    yield ('BL', _0_707)
                # BC
                if in_src('BC') and not in_dst('BC'):
                    yield ('BC', _0_500)
            # mix top
            if not in_dst('TFL') and not in_dst('TML') and not in_dst('TBL'):
                if in_src('TFL'):
                    yield ('TFL', _0_707)
                if in_src('TML') and not in_dst('SL'):
                    if in_dst('BL'):
                        yield ('TML', _0_500)
                    else:
                        yield ('TML', _0_707)
                if in_src('TBL') and not in_dst('BL'):
                    yield ('TBL', _0_500)
        case 'FR':
            # symmetrical to FL
            yield from r_from_l()
        case 'FLC':
            # mix front
            if in_src('FLC'):
                yield ('FLC', _1_000)
            else:
                if in_src('FL'):
                    yield ('FL', _0_707)
                if in_src('FC'):
                    yield ('FC', _0_500)
            # mix surround and top as we would mix into FL
            for chan, coeff in map_chan('FL', dst_mask, src_mask):
                if is_l(chan) and (is_s(chan) or is_b(chan) or is_t(chan)):
                    yield (chan, coeff)
        case 'FRC':
            # symmetrical to FLC
            yield from r_from_l()
        case 'FC':
            # mix front
            if in_src('FL') and (not in_dst('FL') or not in_src('FC') or
                                 (in_src('BC') and not in_dst('BL'))):
                yield ('FL', _0_707)
            if in_src('FLC') and not in_dst('FLC'):
                yield ('FLC', _0_707)
            if in_src('FC'):
                yield ('FC', _1_000)
            if in_src('FRC') and not in_dst('FRC'):
                yield ('FRC', _0_707)
            if in_src('FR') and (not in_dst('FR') or not in_src('FC') or
                                 (in_src('BC') and not in_dst('BR'))):
                yield ('FR', _0_707)
            # mix surround
            if in_src('SL') and not in_dst('SL') and not in_dst('BL') and (
                    not in_dst('FL') or not in_src('FC')):
                yield ('SL', _0_500)
            if in_src('SR') and not in_dst('SR') and not in_dst('BR') and (
                    not in_dst('FR') or not in_src('FC')):
                yield ('SR', _0_500)
            if in_src('BL') and not in_dst('BL') and (
                    not in_dst('FL') or not in_src('FC') or in_src('BC')):
                yield ('BL', _0_500)
            if in_src('BC') and not in_dst('BC') and (
                    not in_dst('BL') and not in_dst('BR') and
                    not in_dst('FL') and not in_dst('FR')):
                yield ('BC', _0_707)
            if in_src('BR') and not in_dst('BR') and (
                    not in_dst('FR') or not in_src('FC') or in_src('BC')):
                yield ('BR', _0_500)
            # mix top
            if in_src('TFL') and not in_dst('TFL') and not in_dst('FL'):
                yield ('TFL', _0_500)
            if in_src('TFR') and not in_dst('TFR') and not in_dst('FR'):
                yield ('TFR', _0_500)
            if in_src('TML') and not in_dst('TML') and not in_dst('FL'):
                yield ('TML', _0_500)
            if in_src('TMR') and not in_dst('TMR') and not in_dst('FR'):
                yield ('TMR', _0_500)
            if in_src('TBL') and not in_dst('TBL') and not in_dst('FL'):
                yield ('TBL', _0_354)
            if in_src('TBR') and not in_dst('TBR') and not in_dst('FR'):
                yield ('TBR', _0_354)
        case 'SL':
            # mix surround
            if in_src('SL'):
                yield ('SL', _1_000)
            # mix top
            if in_src('TML') and not in_dst('TML'):
                yield ('TML', _0_707)
        case 'SR':
            # symmetrical to SL
            yield from r_from_l()
        case 'BL':
            # mix surround
            if in_src('SL') and not in_dst('SL'):
                yield ('SL', _1_000)
            if in_src('BL'):
                yield ('BL', _1_000)
            if in_src('BC') and not in_dst('BC'):
                yield ('BC', _0_707)
            # mix top
            if not in_dst('TFL') and not in_dst('TML') and not in_dst('TBL'):
                if in_src('TML') and not in_dst('SL'):
                    if in_dst('FL'):
                        yield ('TML', _0_500)
                    else:
                        yield ('TML', _0_707)
                if in_src('TBL') and not in_dst('TBL'):
                    yield ('TBL', _0_707)
        case 'BR':
            # symmetrical to BL
            yield from r_from_l()
        case 'BC':
            # mix surround
            if in_src('BL') and (not in_dst('BL') or not in_src('BC')):
                yield ('BL', _1_000)
            if in_src('BC'):
                yield ('BC', _1_000)
            if in_src('BR') and (not in_dst('BR') or not in_src('BC')):
                yield ('BR', _1_000)
            if in_src('SL') and not in_dst('SL'):
                yield ('SL', _1_000)
            if in_src('SR') and not in_dst('SR'):
                yield ('SR', _1_000)
            # mix top
            if in_src('TML') and not in_dst('TML') and not in_dst('TBL'):
                yield ('TML', _0_707)
            if in_src('TMR') and not in_dst('TMR') and not in_dst('TBR'):
                yield ('TMR', _0_707)
            if in_src('TBL') and not in_dst('TBL'):
                yield ('TBL', _0_707)
            if in_src('TBR') and not in_dst('TBR'):
                yield ('TBR', _0_707)
        case 'TML':
            # mix top
            if in_src('TML'):
                yield ('TML', _1_000)
            if in_src('TFL') and not in_dst('TFL'):
                yield ('TFL', _0_707)
            if in_src('TBL') and not in_dst('TBL'):
                yield ('TBL', _0_707)
            # mix front & back
            if not in_src('TML') and not in_src('TFL') and not in_src('TBL'):
                if in_src('FL'):
                    yield ('FL', _1_000)
                if in_src('BL'):
                    yield ('BL', _1_000)
                if in_src('BC'):
                    yield ('BC', _0_707)
        case 'TMR':
            # symmetrical to TML
            yield from r_from_l()
        case 'TFL':
            # mix top
            if in_src('TFL'):
                yield ('TFL', _1_000)
            if in_src('TML') and not in_dst('TML'):
                yield ('TML', _0_707)
                if in_src('TBL') and not in_dst('TBL'):
                    yield ('TBL', _0_500)
            elif in_src('TBL') and not in_dst('TBL'):
                yield ('TBL', _0_707)
            # mix front
            if not in_src('TML') and not in_src('TFL') and not in_src('TBL'):
                if in_src('FL'):
                    yield ('FL', _1_000)
        case 'TFR':
            # symmetrical to TFR
            yield from r_from_l()
        case 'TBL':
            # mix top
            if in_src('TBL'):
                yield ('TBL', _1_000)
            if in_src('TML') and not in_dst('TML'):
                yield ('TML', _0_707)
                if in_src('TFL') and not in_dst('TFL'):
                    yield ('TFL', _0_500)
            elif in_src('TFL') and not in_dst('TFL'):
                yield ('TBL', _0_707)
            # mix back
            if not in_src('TML') and not in_src('TFL') and not in_src('TBL'):
                if in_src('BL'):
                    yield ('BL', _1_000)
                if in_src('BC'):
                    yield ('BC', _0_707)
        case 'TBR':
            # symmetrical to TBR
            yield from r_from_l()
        case 'LFE':
            # LFE is just copied
            if in_src('LFE'):
                yield ('LFE', _1_000)
        case _:
            raise NotImplementedError(dst_chan)

env = jinja2.Environment(
    trim_blocks=True,
    lstrip_blocks=True,
    undefined = jinja2.StrictUndefined)

template = env.from_string('''
/*
 * THIS FILE IS AUTO-GENERATED USING `channel_tables_gen.py'. DO NOT EDIT!
 */

#include "roc_audio/channel_tables.h"
#include "roc_core/macro_helpers.h"

namespace roc {
namespace audio {

// Table of channel position names.
const ChannelPositionName ChanPositionNames[{{ len(chan_pos_names()) }}] = {
{% for chan_id, chan_name in chan_pos_names() %}
    { "{{ chan_name }}", ChanPos_{{ chan_id }} },
{% endfor %}
};

// Table of channel mask names.
const ChannelMaskName ChanMaskNames[{{ len(chan_mask_names()) }}] = {
{% for mask_id, mask_name in chan_mask_names() %}
    { "{{ mask_name }}", ChanMask_Surround_{{ mask_id }} },
{% endfor %}
};

// Table of channel orders.
const ChannelOrderTable ChanOrderTables[{{ len(chan_orders()) }}] = {
{% for order_id, order_name, order_chans in chan_orders() %}
    {
        "{{ order_name }}",
        ChanOrder_{{ order_id }},
        {
{% for ch in order_chans %}
            ChanPos_{{ CHANS[ch] }},
{% endfor %}
            ChanPos_Max,
        },
    },
{% endfor %}
};

// Table of channel mappings.
const ChannelMapTable ChanMapTables[{{ len(chan_mappings()) }}] = {
{% for src_start, src_name, src_mask, dst_name, dst_mask in chan_mappings() %}
{% if src_start %}
    // {{ src_name }}<>...
{% endif %}
    {
        "{{ src_name }}<>{{ dst_name }}",
        ChanMask_Surround_{{ src_mask.id }},
        ChanMask_Surround_{{ dst_mask.id }},
        {
{% for dst_chan in dst_mask.chans %}
            // {{ dst_chan }}
{% for src_chan, coeff in map_chan(dst_chan, dst_mask.chans, src_mask.chans) %}
            { ChanPos_{{ CHANS[dst_chan] }}, ChanPos_{{ CHANS[src_chan] }}, {{ "%.7f"|format(coeff) }}f },
{% endfor %}
{% endfor %}
        },
    },
{% endfor %}
};

} // namespace audio
} // namespace roc
'''.strip())

text = template.render(
    **dict(list(globals().items()) + list(builtins.__dict__.items())),
    )

os.chdir(os.path.dirname(os.path.abspath(__file__)))

with open('channel_tables.cpp', 'w') as fp:
    print(text, file=fp)



#! /usr/bin/env python3

def print_header():
    print('''
/*
 * THIS FILE IS AUTO-GENERATED USING `atomic_ops_gen.py'.
 */

#ifndef ROC_CORE_ATOMIC_OPS_H_
#define ROC_CORE_ATOMIC_OPS_H_

#include "roc_core/stddefs.h"

#include <atomic_ops.h>

namespace roc {
namespace core {

//! Atomic operations.
class AtomicOps {
public:
    #ifdef AO_HAVE_nop_read
    //! Acquire memory barrier.
    static inline void fence_acquire() {
        AO_nop_read();
    }
    #endif // AO_HAVE_nop_read

    #ifdef AO_HAVE_nop_write
    //! Release memory barrier.
    static inline void fence_release() {
        AO_nop_write();
    }
    #endif // AO_HAVE_nop_write

    #ifdef AO_HAVE_nop_full
    //! Full memory barrier.
    static inline void fence_seq_cst() {
        AO_nop_full();
    }
    #endif // AO_HAVE_nop_full
'''.strip())

def print_load(**keys):
    print('''
    #ifdef AO_HAVE{family}_load{fence_impl}
    //! Atomic load ({fence_name} barrier).
    {qualifier} {type_func} load_{fence_func}({type_func} const& var) {{
        struct TypeCheck {{
            int f : sizeof({type_func}) == sizeof({type_impl}) ? 1 : -1;
        }};
        return ({type_func})AO{family}_load{fence_impl}(({type_impl} const*)&var);
    }}
    #endif // AO_HAVE{family}_load{fence_impl}
'''.rstrip().format(**keys))

def print_store(**keys):
    print('''
    #ifdef AO_HAVE{family}_store{fence_impl}
    //! Atomic store ({fence_name} barrier).
    {qualifier} void store_{fence_func}({type_func}& var, {type_func} val) {{
        struct TypeCheck {{
            int f : sizeof({type_func}) == sizeof({type_impl}) ? 1 : -1;
        }};
        AO{family}_store{fence_impl}(({type_impl}*)&var, ({type_impl})val);
    }}
    #endif // AO_HAVE{family}_store{fence_impl}
'''.rstrip().format(**keys))

def print_exchange(**keys):
    print('''
    #ifdef AO_HAVE{family}_fetch_compare_and_swap{fence_impl}
    //! Atomic exchange ({fence_name} barrier).
    {qualifier} {type_func} exchange_{fence_func}({type_func}& var, {type_func} val) {{
        struct TypeCheck {{
            int f : sizeof({type_func}) == sizeof({type_impl}) ? 1 : -1;
        }};
        {type_impl} curr = AO{family}_load(({type_impl}*)&var);
        {type_impl} prev;
        do {{
            prev = curr;
            curr = AO{family}_fetch_compare_and_swap{fence_impl}(({type_impl}*)&var, prev,
                                                         ({type_impl})val);
        }} while (curr != prev);
        return ({type_func})curr;
    }}
    #endif // AO_HAVE{family}_fetch_compare_and_swap{fence_impl}
'''.rstrip().format(**keys))

def print_exchange_explicit_fence(**keys):
    print('''
    #ifdef AO_HAVE{family}_fetch_compare_and_swap
    //! Atomic exchange ({fence_name} barrier).
    {qualifier} {type_func} exchange_{fence_func}({type_func}& var, {type_func} val) {{
        struct TypeCheck {{
            int f : sizeof({type_func}) == sizeof({type_impl}) ? 1 : -1;
        }};
        AO_nop{fence_impl}();
        {type_impl} curr = AO{family}_load(({type_impl}*)&var);
        {type_impl} prev;
        do {{
            prev = curr;
            curr = AO{family}_fetch_compare_and_swap(({type_impl}*)&var, prev,
                                                         ({type_impl})val);
            AO_nop{fence_impl}();
        }} while (curr != prev);
        return ({type_func})curr;
    }}
    #endif // AO_HAVE{family}_fetch_compare_and_swap
'''.rstrip().format(**keys))

def print_compare_exchange(**keys):
    print('''
    #ifdef AO_HAVE{family}_fetch_compare_and_swap{fence_impl}
    //! Atomic compare-and-swap ({fence_name} barrier).
    {qualifier} bool compare_exchange_{fence_func}(
          {type_func}& var, {type_func}& exp, {type_func} des) {{
        struct TypeCheck {{
            int f : sizeof({type_func}) == sizeof({type_impl}) ? 1 : -1;
        }};
        {type_impl} old = AO{family}_fetch_compare_and_swap{fence_impl}(
          ({type_impl}*)&var, ({type_impl})exp, ({type_impl})des);
        const bool ret = (({type_impl})exp == old);
        exp = ({type_func})old;
        return ret;
    }}
    #endif // AO_HAVE{family}_fetch_compare_and_swap{fence_impl}
'''.rstrip().format(**keys))

def print_compare_exchange_explicit_fence(**keys):
    print('''
    #ifdef AO_HAVE{family}_fetch_compare_and_swap
    //! Atomic compare-and-swap ({fence_name} barrier).
    {qualifier} bool compare_exchange_{fence_func}(
          {type_func}& var, {type_func}& exp, {type_func} des) {{
        struct TypeCheck {{
            int f : sizeof({type_func}) == sizeof({type_impl}) ? 1 : -1;
        }};
        AO_nop{fence_impl}();
        {type_impl} old = AO{family}_fetch_compare_and_swap(
          ({type_impl}*)&var, ({type_impl})exp, ({type_impl})des);
        const bool ret = (({type_impl})exp == old);
        if (ret) {{
            AO_nop{fence_impl}();
        }}
        exp = ({type_func})old;
        return ret;
    }}
    #endif // AO_HAVE{family}_fetch_compare_and_swap
'''.rstrip().format(**keys))

def print_fetch_add(**keys):
    print('''
    #ifdef AO_HAVE{family}_fetch_and_add{fence_impl}
    //! Atomic fetch-add ({fence_name} barrier).
    {qualifier} {type_func} fetch_add_{fence_func}({type_func}& var, {type_arg} val) {{
        struct TypeCheck {{
            int f : sizeof({type_func}) == sizeof({type_impl}) ? 1 : -1;
        }};
        return ({type_func})(
            AO{family}_fetch_and_add{fence_impl}(({type_impl}*)&var, ({type_impl})val));
    }}
    #endif // AO_HAVE{family}_fetch_and_add{fence_impl}
'''.rstrip().format(**keys))

def print_fetch_sub(**keys):
    print('''
    #ifdef AO_HAVE{family}_fetch_and_add{fence_impl}
    //! Atomic fetch-sub ({fence_name} barrier).
    {qualifier} {type_func} fetch_sub_{fence_func}({type_func}& var, {type_arg} val) {{
        struct TypeCheck {{
            int f : sizeof({type_func}) == sizeof({type_impl}) ? 1 : -1;
        }};
        return ({type_func})(
            AO{family}_fetch_and_add{fence_impl}(({type_impl}*)&var, ({type_impl})-val));
    }}
    #endif // AO_HAVE{family}_fetch_and_add{fence_impl}
'''.rstrip().format(**keys))

def print_fetch_bitop(**keys):
    print('''
    #ifdef AO_HAVE{family}_fetch_compare_and_swap{fence_impl}
    //! Atomic fetch-{bitop_name} ({fence_name} barrier).
    {qualifier} {type_func} fetch_{bitop_name}_{fence_func}({type_func}& var, {type_arg} val) {{
        struct TypeCheck {{
            int f : sizeof({type_func}) == sizeof({type_impl}) ? 1 : -1;
        }};
        {type_impl} curr = AO{family}_load(({type_impl}*)&var);
        {type_impl} prev;
        do {{
            prev = curr;
            curr = AO{family}_fetch_compare_and_swap{fence_impl}(({type_impl}*)&var, prev,
                                                         prev {bitop_sign} ({type_impl})val);
        }} while (curr != prev);
        return ({type_func})curr;
    }}
    #endif // AO_HAVE{family}_fetch_compare_and_swap{fence_impl}
'''.rstrip().format(**keys))

def print_fetch_bitop_explicit_fence(**keys):
    print('''
    #ifdef AO_HAVE{family}_fetch_compare_and_swap{fence_impl}
    //! Atomic fetch-{bitop_name} ({fence_name} barrier).
    {qualifier} {type_func} fetch_{bitop_name}_{fence_func}({type_func}& var, {type_arg} val) {{
        struct TypeCheck {{
            int f : sizeof({type_func}) == sizeof({type_impl}) ? 1 : -1;
        }};
        AO_nop{fence_impl}();
        {type_impl} curr = AO{family}_load(({type_impl}*)&var);
        {type_impl} prev;
        do {{
            prev = curr;
            curr = AO{family}_fetch_compare_and_swap(({type_impl}*)&var, prev,
                                                         prev {bitop_sign} ({type_impl})val);
            AO_nop{fence_impl}();
        }} while (curr != prev);
        return ({type_func})curr;
    }}
    #endif // AO_HAVE{family}_fetch_compare_and_swap{fence_impl}
'''.rstrip().format(**keys))

def print_methods(**kw):
    print('''
    // overloads for {type_func}
'''.rstrip().format(**kw))

    kw['fence_name'] = 'no'
    kw['fence_func'] = 'relaxed'
    kw['fence_impl'] = ''
    print_load(**kw)
    print_store(**kw)
    print_exchange(**kw)
    print_compare_exchange(**kw)
    print_fetch_add(**kw)
    print_fetch_sub(**kw)
    print_fetch_bitop(bitop_name='and', bitop_sign='&', **kw)
    print_fetch_bitop(bitop_name='or', bitop_sign='|', **kw)
    print_fetch_bitop(bitop_name='xor', bitop_sign='^', **kw)

    kw['fence_name'] = 'acquire'
    kw['fence_func'] = 'acquire'
    kw['fence_impl'] = '_acquire'
    print_load(**kw)
    print_exchange(**kw)
    print_compare_exchange(**kw)
    print_fetch_add(**kw)
    print_fetch_sub(**kw)
    print_fetch_bitop(bitop_name='and', bitop_sign='&', **kw)
    print_fetch_bitop(bitop_name='or', bitop_sign='|', **kw)
    print_fetch_bitop(bitop_name='xor', bitop_sign='^', **kw)

    kw['fence_name'] = 'release'
    kw['fence_func'] = 'release'
    kw['fence_impl'] = '_release'
    print_store(**kw)
    print_exchange(**kw)
    print_compare_exchange(**kw)
    print_fetch_add(**kw)
    print_fetch_sub(**kw)
    print_fetch_bitop(bitop_name='and', bitop_sign='&', **kw)
    print_fetch_bitop(bitop_name='or', bitop_sign='|', **kw)
    print_fetch_bitop(bitop_name='xor', bitop_sign='^', **kw)

    kw['fence_name'] = 'acquire-release'
    kw['fence_func'] = 'acq_rel'
    kw['fence_impl'] = '_full'
    print_exchange(**kw)
    print_compare_exchange(**kw)

    kw['fence_name'] = 'full'
    kw['fence_func'] = 'seq_cst'
    kw['fence_impl'] = '_full'
    print_load(**kw)
    print_store(**kw)
    print_exchange_explicit_fence(**kw)
    print_compare_exchange_explicit_fence(**kw)
    print_fetch_add(**kw)
    print_fetch_sub(**kw)
    print_fetch_bitop_explicit_fence(bitop_name='and', bitop_sign='&', **kw)
    print_fetch_bitop_explicit_fence(bitop_name='or', bitop_sign='|', **kw)
    print_fetch_bitop_explicit_fence(bitop_name='xor', bitop_sign='^', **kw)

def print_footer():
    print('''
};

} // namespace core
} // namespace roc

#endif // ROC_CORE_ATOMIC_OPS_H_
'''.strip())

print_header()

print_methods(
    qualifier='static inline',
    family='_char',
    type_func='unsigned char',
    type_arg='unsigned char',
    type_impl='unsigned char')

print_methods(
    qualifier='static inline',
    family='_char',
    type_func='char',
    type_arg='char',
    type_impl='unsigned char')

print_methods(
    qualifier='static inline',
    family='_short',
    type_func='unsigned short',
    type_arg='unsigned short',
    type_impl='unsigned short')

print_methods(
    qualifier='static inline',
    family='_short',
    type_func='short',
    type_arg='short',
    type_impl='unsigned short')

print()
print('#ifndef AO_T_IS_INT')

print_methods(
    qualifier='static inline',
    family='_int',
    type_func='unsigned int',
    type_arg='unsigned int',
    type_impl='unsigned int')

print_methods(
    qualifier='static inline',
    family='_int',
    type_func='int',
    type_arg='int',
    type_impl='unsigned int')

print()
print('#endif // AO_T_IS_INT')

print_methods(
    qualifier='static inline',
    family='',
    type_func='size_t',
    type_arg='size_t',
    type_impl='AO_t')

print_methods(
    qualifier='static inline',
    family='',
    type_func='ssize_t',
    type_arg='ssize_t',
    type_impl='AO_t')

print_methods(
    qualifier='template <class T> static inline',
    family='',
    type_func='T*',
    type_arg='ptrdiff_t',
    type_impl='AO_t')

print_footer()



import SCons.Script
import os

def _find_argument(env, arg):
    ret = SCons.Script.ARGUMENTS.get(arg, None)
    if ret is None:
        if arg in os.environ:
            ret = os.environ[arg]
    return ret

def _mark_known_argument(env, arg):
    if '_KNOWN_ARGS' not in env.Dictionary():
        env['_KNOWN_ARGS'] = set()
    env['_KNOWN_ARGS'].add(arg)

def _is_known_argument(env, arg):
    if '_KNOWN_ARGS' in env.Dictionary():
        return arg in env['_KNOWN_ARGS']
    return False

def HasArgument(env, arg):
    return _is_known_argument(env, arg) or _find_argument(env, arg) is not None

def OverrideFromArgument(env, arg, names=[], default=None):
    if not names:
        names = [arg]
    for name in names:
        v = _find_argument(env, name)
        if v is not None:
            _mark_known_argument(env, arg)
        else:
            v = default
        if v is not None:
            env[arg] = v
            break

def PrependFromArgument(env, arg, names=[], default=None):
    if not names:
        names = [arg]
    for name in names:
        v = _find_argument(env, name)
        if v is not None:
            _mark_known_argument(env, arg)
        else:
            v = default
        if v is not None:
            env.Prepend(**{arg: v})
            break

def init(env):
    env.AddMethod(HasArgument, 'HasArgument')
    env.AddMethod(OverrideFromArgument, 'OverrideFromArgument')
    env.AddMethod(PrependFromArgument, 'PrependFromArgument')



#! /usr/bin/env python3
import glob
import json
import os.path
import re
import sys

def print_comment(key, indent=1):
    print("%s/* %-16s */ " % (" " * indent * 2, key), end="")

def print_uint(key, value):
    print_comment(key)
    print("%su," % value)

def print_bool(key, value):
    print_comment(key)
    print("%s," % ("1" if value else "0"))

def print_array(key, array, indent=1):
    if len(array) > 12:
        ret = "{\n"
        while array:
            ret += " " * (indent * 2 + 1)
            for n in range(12):
                if array:
                    ret += " %s," % array.pop(0)
            ret += "\n"
        ret += " " * (indent * 2)
        ret += "},"
    else:
        ret = "{ "
        ret += ", ".join(map(str, array))
        ret += " },"

    print_comment(key, indent)
    print(ret)

def print_blob(key, blob):
    print_array(key, ["0x%02x" % (ord(b) if isinstance(b, str) else b) for b in blob[:]])

if len(sys.argv) != 1:
    print("usage: generate_headers.py")
    exit(1)

os.chdir(os.path.dirname(os.path.abspath(__file__)))

for in_json in glob.glob("*.json"):
    in_name = re.sub("\.json$", "", in_json)
    in_blob = in_name + ".blob"

    if not os.path.exists(in_blob):
        print("%s doesn't exist, skipping %s" % (in_blob, in_json), file=sys.stderr)
        continue

    out_h = in_name + ".h"
    print("writing %s" % out_h, file=sys.stderr)

    guard = 'ROC_RTP_TEST_PACKETS_' + out_h.upper().replace('.', '_') + '_'

    meta = json.loads(open(in_json).read())

    with open(in_blob, 'rb') as blob:
        raw_data = blob.read()

    with open(out_h, 'w') as sys.stdout:
        print(f"""
/*
 * THIS FILE IS AUTO-GENERATED USING `generate_headers.py'
 *
 * Inputs:
 *  - {in_json}
 *  - {in_blob}
 */

#ifndef {guard}
#define {guard}

#include "test_packets/packet_info.h"

namespace roc {{
namespace rtp {{
namespace test {{

static PacketInfo {in_name} = {{
        """.strip())

        print_blob("raw_data", raw_data)

        print()
        print_uint("packet_size", len(raw_data))
        print_uint("header_size", meta["header_size"])
        print_uint("extension_size", meta["extension_size"])
        print_uint("payload_size", meta["payload_size"])
        print_uint("padding_size", meta["padding_size"])

        print()
        print_uint("version", meta["version"])
        print_bool("padding", meta["padding"])
        print_bool("extension", meta["extension"])
        print_uint("num_csrc", len(meta["csrc"]))
        print_uint("pt", meta["pt"])
        print_bool("marker", meta["marker"])

        print()
        print_uint("seqnum", meta["seqnum"])
        print_uint("ts", meta["ts"])
        print_uint("ssrc", meta["ssrc"])
        print_array("csrc", meta["csrc"])

        print()
        print_uint("ext_type", meta["ext_type"])
        print_uint("ext_data_size", len(meta["ext_data"]))
        print_blob("ext_data", meta["ext_data"])

        print()
        print_uint("num_channels", meta["num_channels"])
        print_uint("num_samples", meta["num_samples"])
        print_uint("samplebits", meta["samplebits"])
        print_uint("samplerate", meta["samplerate"])

        print()
        print_comment("samples")
        print("{")
        n_ch = 0
        for ch in meta["samples"]:
            print_array("channel #%s" % n_ch, ch, indent=2)
            n_ch += 1
        print("  },")

        print(f"""
}};

}} // namespace test
}} // namespace rtp
}} // namespace roc

#endif // {guard}
        """.strip())



import SCons.Script
import sys
import re

_COLORS = {}
_COMPACT = False

def _init_colors():
    global _COLORS

    if sys.stdout.isatty():
        _COLORS['cyan']   = '\033[0;36m'
        _COLORS['purple'] = '\033[0;35m'
        _COLORS['blue']   = '\033[0;34m'
        _COLORS['green']  = '\033[0;32m'
        _COLORS['yellow'] = '\033[0;33m'
        _COLORS['red']    = '\033[0;31m'
        _COLORS['end']    = '\033[0m'
    else:
        for k in ['cyan', 'purple', 'blue', 'green', 'yellow', 'red', 'end']:
           _COLORS[k] = ''

def _init_compact():
    global _COMPACT

    for arg in sys.argv:
        if re.match('^-([^-].*)?Q.*$', arg):
            _COMPACT = True
            break

def _init_pretty(env):
    global _COMPACT

    if not _COMPACT:
        return

    def _subst(env, string, raw=1, target=None, source=None, conv=None, executor=None):
        raw = 1
        gvars = env.gvars()
        lvars = env.lvars()
        lvars['__env__'] = env
        if executor:
            lvars.update(executor.get_lvars())
        return SCons.Subst.scons_subst(string, env, raw, target, source, gvars, lvars, conv)

    env.AddMethod(_subst, 'subst_target_source')

    env['CCCOMSTR']       = env.PrettyCommand('CC', '$SOURCE', 'blue')
    env['SHCCCOMSTR']     = env.PrettyCommand('CC', '$SOURCE', 'blue')

    env['CXXCOMSTR']      = env.PrettyCommand('CXX', '$SOURCE', 'blue')
    env['SHCXXCOMSTR']    = env.PrettyCommand('CXX', '$SOURCE', 'blue')

    env['RCCOMSTR']       = env.PrettyCommand('RC', '$SOURCE', 'red')
    env['ARCOMSTR']       = env.PrettyCommand('AR', '$TARGET', 'red')
    env['SHLINKCOMSTR']   = env.PrettyCommand('LD', '$TARGET', 'red')
    env['LDMODULECOMSTR'] = env.PrettyCommand('LD', '$TARGET', 'red')
    env['LINKCOMSTR']     = env.PrettyCommand('LD', '$TARGET', 'red')
    env['RANLIBCOMSTR']   = env.PrettyCommand('RANLIB', '$TARGET', 'red')
    env['MTCOMSTR']       = env.PrettyCommand('MT', '$TARGET', 'green')

    env['INSTALLSTR']     = env.PrettyCommand('INSTALL', '$TARGET', 'yellow')

def PrettyCommand(env, command, args, color, cmdline=None):
    global _COMPACT

    if _COMPACT:
        return ' {}{:>8s}{}   {}'.format(_COLORS[color], command, _COLORS['end'], args)
    elif cmdline:
        return cmdline
    else:
        return '$CMDLINE'

def init(env):
    env.AddMethod(PrettyCommand, 'PrettyCommand')
    _init_colors()
    _init_compact()
    _init_pretty(env)



#! /usr/bin/env python3
import builtins
import math
import os
import os.path
import sys

try:
    import jinja2
except ImportError:
    print('''
error: can't import python module "jinja2", install it using "pip3 install Jinja2"
'''.strip(), file=sys.stderr)
    exit(1)

# generate initializers (C expressions) for min and max
# values for given pcm code
#
# e.g. for sint8:
#  min: -127 - 1
#  max: 127
def compute_minmax(code):
    suffix = ''
    if not code['is_signed']:
        suffix += 'u'
    if code['width'] >= 32:
        suffix += 'l'
    if code['width'] >= 64:
        suffix += 'l'

    if code['is_signed']:
        power = pow(2, code['width']-1) - 1

        min_val = f'-{power}{suffix} - 1'
        max_val = f'{power}{suffix}'
    else:
        power = pow(2, code['width']) - 1

        min_val = f'0{suffix}'
        max_val = f'{power}{suffix}'

    return min_val, max_val

# generate masks (C expressions) for given pcm code:
#  value_mask: masks all significant bits
#  sign_mask: masks sign bit
#  lsb_mask: masks all bits to the left of the sign bit
#
# e.g. for sint18_3:
#  value_masks: masks first 18 bits (0x3ffff)
#  sign_mask: masks 18th bit (0x20000)
#  lsb_mask: masks bits after 18th (0xfffc0000)
def compute_masks(code):
    value_mask = hex(int('1' * code['width'], 2))
    if not code['is_signed']:
        value_mask += 'u'

    if code['is_signed']:
        sign_mask = hex(int('1' + '0' * (code['width']-1), 2))
    else:
        sign_mask = None

    if code['is_signed'] and code['unpacked_width'] > code['width']:
        lsb_mask = hex(
            int(
                ('1' * (code['unpacked_width'] - code['width']) +
                '0' * code['width']),
                2))
    else:
        lsb_mask = None

    return value_mask, sign_mask, lsb_mask

# compute octet counts for given pcm code:
#  significant_octets: number of octets needed to fit all significant bits
#  packed_octets: number of octets needed to fit packed value
#  unpacked_octets: number of octets needed to fit unpacked value
#
# e.g. for sint18_3:
#  significant_octets: 3
#  packed_octets: 3
#  unpacked_octets: 4
def compute_octets(code):
    significant_octets = math.ceil(code['width'] / 8)
    packed_octets = math.ceil(code['packed_width'] / 8)
    unpacked_octets = math.ceil(code['unpacked_width'] / 8)

    return significant_octets, packed_octets, unpacked_octets

# generate enum name for pcm code + endian
# e.g.:
#  SInt18_3, Native => PcmFormat_SInt18_3
#  SInt18_3, Little => PcmFormat_SInt18_3_Le
#  SInt18_3, Big => PcmFormat_SInt18_3_Be
def make_enum_name(code, endian):
    name = code['code']

    if endian != 'Native':
        if endian == 'Little':
            name += '_Le'
        if endian == 'Big':
            name += '_Be'

    return 'PcmFormat_' + name

# generate string name for pcm code + endian
# e.g.:
#  SInt18_3, Native => s18_3
#  SInt18_3, Little => s18_3le
#  SInt18_3, Big => s18_3be
def make_str_name(code, endian):
    name = code['short_name']

    if endian != 'Native':
        if not '_' in name:
            name += '_'
        if endian == 'Little':
            name += 'le'
        if endian == 'Big':
            name += 'be'

    return name

# find all codes which short name has given prefix, and return
# sorted list of all possible characters in short name that can
# follow next to the prefix
#
# e.g. if prefix is ('f',), then codes starting with prefix
# are f32* and f64*, and nth_chars() returns ['3', '6']
def nth_chars(codes, prefix=()):
    pos = len(prefix)
    ret = set()
    for code in codes:
        name = code['short_name']
        if len(name) > pos and tuple(name[:pos]) == prefix:
            ret.add(name[pos])
    return list(sorted(ret))

CODES = [
    {
        'code': 'SInt8',
        'signed_code': 'SInt8',
        'type': 'int8_t',
        'signed_type': 'int8_t',
        'unsigned_type': 'uint8_t',
        'short_name': 's8',
        'is_integer': True,
        'is_signed': True,
        'depth': 8,
        'width': 8,
        'packed_width': 8,
        'unpacked_width': 8,
    },
    {
        'code': 'UInt8',
        'signed_code': 'SInt8',
        'type': 'uint8_t',
        'signed_type': 'int8_t',
        'unsigned_type': 'uint8_t',
        'short_name': 'u8',
        'is_integer': True,
        'is_signed': False,
        'depth': 8,
        'width': 8,
        'packed_width': 8,
        'unpacked_width': 8,
    },
    {
        'code': 'SInt16',
        'signed_code': 'SInt16',
        'type': 'int16_t',
        'signed_type': 'int16_t',
        'unsigned_type': 'uint16_t',
        'short_name': 's16',
        'is_integer': True,
        'is_signed': True,
        'depth': 16,
        'width': 16,
        'packed_width': 16,
        'unpacked_width': 16,
    },
    {
        'code': 'UInt16',
        'signed_code': 'SInt16',
        'type': 'uint16_t',
        'signed_type': 'int16_t',
        'unsigned_type': 'uint16_t',
        'short_name': 'u16',
        'is_integer': True,
        'is_signed': False,
        'depth': 16,
        'width': 16,
        'packed_width': 16,
        'unpacked_width': 16,
    },
    {
        'code': 'SInt18',
        'signed_code': 'SInt18',
        'type': 'int32_t',
        'signed_type': 'int32_t',
        'unsigned_type': 'uint32_t',
        'short_name': 's18',
        'is_integer': True,
        'is_signed': True,
        'depth': 18,
        'width': 18,
        'packed_width': 18,
        'unpacked_width': 32,
    },
    {
        'code': 'UInt18',
        'signed_code': 'SInt18',
        'type': 'uint32_t',
        'signed_type': 'int32_t',
        'unsigned_type': 'uint32_t',
        'short_name': 'u18',
        'is_integer': True,
        'is_signed': False,
        'depth': 18,
        'width': 18,
        'packed_width': 18,
        'unpacked_width': 32,
    },
    {
        'code': 'SInt18_3',
        'signed_code': 'SInt18_3',
        'type': 'int32_t',
        'signed_type': 'int32_t',
        'unsigned_type': 'uint32_t',
        'short_name': 's18_3',
        'is_integer': True,
        'is_signed': True,
        'depth': 18,
        'width': 18,
        'packed_width': 24,
        'unpacked_width': 32,
    },
    {
        'code': 'UInt18_3',
        'signed_code': 'SInt18_3',
        'type': 'uint32_t',
        'signed_type': 'int32_t',
        'unsigned_type': 'uint32_t',
        'short_name': 'u18_3',
        'is_integer': True,
        'is_signed': False,
        'depth': 18,
        'width': 18,
        'packed_width': 24,
        'unpacked_width': 32,
    },
    {
        'code': 'SInt18_4',
        'signed_code': 'SInt18_4',
        'type': 'int32_t',
        'signed_type': 'int32_t',
        'unsigned_type': 'uint32_t',
        'short_name': 's18_4',
        'is_integer': True,
        'is_signed': True,
        'depth': 18,
        'width': 18,
        'packed_width': 32,
        'unpacked_width': 32,
    },
    {
        'code': 'UInt18_4',
        'signed_code': 'SInt18_4',
        'type': 'uint32_t',
        'signed_type': 'int32_t',
        'unsigned_type': 'uint32_t',
        'short_name': 'u18_4',
        'is_integer': True,
        'is_signed': False,
        'depth': 18,
        'width': 18,
        'packed_width': 32,
        'unpacked_width': 32,
    },
    {
        'code': 'SInt20',
        'signed_code': 'SInt20',
        'type': 'int32_t',
        'signed_type': 'int32_t',
        'unsigned_type': 'uint32_t',
        'short_name': 's20',
        'is_integer': True,
        'is_signed': True,
        'depth': 20,
        'width': 20,
        'packed_width': 20,
        'unpacked_width': 32,
    },
    {
        'code': 'UInt20',
        'signed_code': 'SInt20',
        'type': 'uint32_t',
        'signed_type': 'int32_t',
        'unsigned_type': 'uint32_t',
        'short_name': 'u20',
        'is_integer': True,
        'is_signed': False,
        'depth': 20,
        'width': 20,
        'packed_width': 20,
        'unpacked_width': 32,
    },
    {
        'code': 'SInt20_3',
        'signed_code': 'SInt20_3',
        'type': 'int32_t',
        'signed_type': 'int32_t',
        'unsigned_type': 'uint32_t',
        'short_name': 's20_3',
        'is_integer': True,
        'is_signed': True,
        'depth': 20,
        'width': 20,
        'packed_width': 24,
        'unpacked_width': 32,
    },
    {
        'code': 'UInt20_3',
        'signed_code': 'SInt20_3',
        'type': 'uint32_t',
        'signed_type': 'int32_t',
        'unsigned_type': 'uint32_t',
        'short_name': 'u20_3',
        'is_integer': True,
        'is_signed': False,
        'depth': 20,
        'width': 20,
        'packed_width': 24,
        'unpacked_width': 32,
    },
    {
        'code': 'SInt20_4',
        'signed_code': 'SInt20_4',
        'type': 'int32_t',
        'signed_type': 'int32_t',
        'unsigned_type': 'uint32_t',
        'short_name': 's20_4',
        'is_integer': True,
        'is_signed': True,
        'depth': 20,
        'width': 20,
        'packed_width': 32,
        'unpacked_width': 32,
    },
    {
        'code': 'UInt20_4',
        'signed_code': 'SInt20_4',
        'type': 'uint32_t',
        'signed_type': 'int32_t',
        'unsigned_type': 'uint32_t',
        'short_name': 'u20_4',
        'is_integer': True,
        'is_signed': False,
        'depth': 20,
        'width': 20,
        'packed_width': 32,
        'unpacked_width': 32,
    },
    {
        'code': 'SInt24',
        'signed_code': 'SInt24',
        'type': 'int32_t',
        'signed_type': 'int32_t',
        'unsigned_type': 'uint32_t',
        'short_name': 's24',
        'is_integer': True,
        'is_signed': True,
        'depth': 24,
        'width': 24,
        'packed_width': 24,
        'unpacked_width': 32,
    },
    {
        'code': 'UInt24',
        'signed_code': 'SInt24',
        'type': 'uint32_t',
        'signed_type': 'int32_t',
        'unsigned_type': 'uint32_t',
        'short_name': 'u24',
        'is_integer': True,
        'is_signed': False,
        'depth': 24,
        'width': 24,
        'packed_width': 24,
        'unpacked_width': 32,
    },
    {
        'code': 'SInt24_4',
        'signed_code': 'SInt24_4',
        'type': 'int32_t',
        'signed_type': 'int32_t',
        'unsigned_type': 'uint32_t',
        'short_name': 's24_4',
        'is_integer': True,
        'is_signed': True,
        'depth': 24,
        'width': 24,
        'packed_width': 32,
        'unpacked_width': 32,
    },
    {
        'code': 'UInt24_4',
        'signed_code': 'SInt24_4',
        'type': 'uint32_t',
        'signed_type': 'int32_t',
        'unsigned_type': 'uint32_t',
        'short_name': 'u24_4',
        'is_integer': True,
        'is_signed': False,
        'depth': 24,
        'width': 24,
        'packed_width': 32,
        'unpacked_width': 32,
    },
    {
        'code': 'SInt32',
        'signed_code': 'SInt32',
        'type': 'int32_t',
        'signed_type': 'int32_t',
        'unsigned_type': 'uint32_t',
        'short_name': 's32',
        'is_integer': True,
        'is_signed': True,
        'depth': 32,
        'width': 32,
        'packed_width': 32,
        'unpacked_width': 32,
    },
    {
        'code': 'UInt32',
        'signed_code': 'SInt32',
        'type': 'uint32_t',
        'signed_type': 'int32_t',
        'unsigned_type': 'uint32_t',
        'short_name': 'u32',
        'is_integer': True,
        'is_signed': False,
        'depth': 32,
        'width': 32,
        'packed_width': 32,
        'unpacked_width': 32,
    },
    {
        'code': 'SInt64',
        'signed_code': 'SInt64',
        'type': 'int64_t',
        'signed_type': 'int64_t',
        'unsigned_type': 'uint64_t',
        'short_name': 's64',
        'is_integer': True,
        'is_signed': True,
        'depth': 64,
        'width': 64,
        'packed_width': 64,
        'unpacked_width': 64,
    },
    {
        'code': 'UInt64',
        'signed_code': 'SInt64',
        'type': 'uint64_t',
        'signed_type': 'int64_t',
        'unsigned_type': 'uint64_t',
        'short_name': 'u64',
        'is_integer': True,
        'is_signed': False,
        'depth': 64,
        'width': 64,
        'packed_width': 64,
        'unpacked_width': 64,
    },
    {
        'code': 'Float32',
        'signed_code': 'Float32',
        'type': 'float',
        'signed_type': 'float',
        'short_name': 'f32',
        'is_integer': False,
        'is_signed': True,
        'depth': 25,
        'width': 32,
        'packed_width': 32,
        'unpacked_width': 32,
    },
    {
        'code': 'Float64',
        'signed_code': 'Float64',
        'type': 'double',
        'signed_type': 'double',
        'short_name': 'f64',
        'is_integer': False,
        'is_signed': True,
        'depth': 53,
        'width': 64,
        'packed_width': 64,
        'unpacked_width': 64,
    },
]

ENDIANS = [
    'Native',
    'Big',
    'Little',
]

TYPES = [
    ('int8_t', 1),
    ('uint8_t', 1),
    ('int16_t', 2),
    ('uint16_t', 2),
    ('int32_t', 4),
    ('uint32_t', 4),
    ('int64_t', 8),
    ('uint64_t', 8),
    ('float', 4),
    ('double', 8),
]

for code in CODES:
    code['min'] = f"pcm_{code['code'].lower()}_min"
    code['max'] = f"pcm_{code['code'].lower()}_max"

    code['signed_min'] = f"pcm_{code['signed_code'].lower()}_min"
    code['signed_max'] = f"pcm_{code['signed_code'].lower()}_max"

    code['min_value'], code['max_value'] = compute_minmax(code)

    code['value_mask'], code['sign_mask'], code['lsb_mask'] = \
      compute_masks(code)

    code['significant_octets'], code['packed_octets'], code['unpacked_octets'] = \
      compute_octets(code)

env = jinja2.Environment(
    trim_blocks=True,
    lstrip_blocks=True,
    undefined = jinja2.StrictUndefined)

template = env.from_string('''
/*
 * THIS FILE IS AUTO-GENERATED USING `pcm_format_gen.py'. DO NOT EDIT!
 */

#include "roc_audio/pcm_format.h"
#include "roc_core/attributes.h"
#include "roc_core/cpu_traits.h"
#include "roc_core/stddefs.h"

namespace roc {
namespace audio {

namespace {

// PCM codes.
enum PcmCode {
{% for code in CODES %}
    PcmCode_{{ code.code }},
{% endfor %}
    PcmCode_Max
};

// PCM endians.
enum PcmEndian {
{% for endian in ENDIANS %}
    PcmEndian_{{ endian }},
{% endfor %}
    PcmEndian_Max
};

{% for code in CODES %}
{% if code.is_integer %}
// {{ code.code }} value range
const {{ code.type }} {{ code.min }} = {{ code.min_value }};
const {{ code.type }} {{ code.max }} = {{ code.max_value }};

{% endif %}
{% endfor %}
// Convert between signed and unsigned samples
template <PcmCode> struct pcm_sign_converter;

{% for code in CODES %}
{% if code.is_integer %}
// Convert {{ code.code }} from/to signed/unsigned
template <> struct pcm_sign_converter<PcmCode_{{ code.code }}> {
{% if code.is_signed %}
    // {{ code.code }} from unsigned value
    static inline {{ code.signed_type }} from_unsigned({{ code.unsigned_type }} arg) {
        if (arg < {{ code.unsigned_type }}({{ code.signed_max }}) + 1) {
            return {{ code.signed_type }}(arg) - {{ code.signed_max }} - 1;
        }
        return {{ code.signed_type }}(arg - {{ code.unsigned_type }}({{ code.signed_max }}) - 1);
    }

    // {{ code.code }} to unsigned value
    static inline {{ code.unsigned_type }} to_unsigned({{ code.signed_type }} arg) {
        if (arg >= 0) {
            return {{ code.unsigned_type }}(arg) + {{ code.signed_max }} + 1;
        }
        return {{ code.unsigned_type }}(arg + {{ code.signed_max }} + 1);
    }
{% else %}
    // {{ code.code }} from signed value
    static inline {{ code.unsigned_type }} from_signed({{ code.signed_type }} arg) {
        if (arg >= 0) {
            return {{ code.unsigned_type }}(arg) + {{ code.signed_max }} + 1;
        }
        return {{ code.unsigned_type }}(arg + {{ code.signed_max }} + 1);
    }

    // {{ code.code }} to signed value
    static inline {{ code.signed_type }} to_signed({{ code.unsigned_type }} arg) {
        if (arg >= {{ code.unsigned_type }}({{ code.signed_max }}) + 1) {
            return {{ code.signed_type }}(arg \
- {{ code.unsigned_type }}({{ code.signed_max }}) - 1);
        }
        return {{ code.signed_type }}(arg - {{ code.unsigned_type }}({{ code.signed_max }}) - 1);
    }
{% endif %}
};

{% endif %}
{% endfor %}
// Convert between unpacked CODES
template <PcmCode InCode, PcmCode OutCode> struct pcm_code_converter;

{% for out in CODES %}
{% for in in CODES %}
{% set both_unsigned = not in.is_signed and not out.is_signed %}
// Convert {{ in.code }} to {{ out.code }}
template <> struct pcm_code_converter<PcmCode_{{ in.code }}, \
PcmCode_{{ out.code }}> {
    static inline {{ out.type }} convert({{ in.type }} arg) {
{% if in.code == out.code %}
        return arg;
{% else %}
{% if not in.is_signed and not both_unsigned %}
        // convert to signed
        {{ in.signed_type }} in = \
pcm_sign_converter<PcmCode_{{ in.code }}>::to_signed(arg);
{% else %}
        {{ in.type }} in = arg;
{% endif %}

{% if both_unsigned %}
        {{ out.type }} out;
{% else %}
        {{ out.signed_type }} out;
{% endif %}
{% if not out.is_integer and not in.is_integer %}
        // float to float
        out = {{ out.type }}(in);
{% elif not out.is_integer and in.is_integer %}
        // integer to float
        out = {{ out.type }}(in * (1.0 / ((double){{ in.signed_max }} + 1.0)));
{% elif out.is_integer and not in.is_integer %}
        // float to integer
        const double d = double(in) * ((double){{ out.signed_max }} + 1.0);
        if (d < {{ out.signed_min }}) {
            // clip
            out = {{ out.signed_min }};
        } else if (d >= (double){{ out.signed_max }} + 1.0) {
            // clip
            out = {{ out.signed_max }};
        } else {
            out = {{ out.signed_type }}(d);
        }
{% elif out.width == in.width %}
        out = in;
{% elif out.width < in.width and both_unsigned %}
        // downscale unsigned integer
        out = {{ out.type }}(in >> {{ in.width - out.width }});
{% elif out.width < in.width and not both_unsigned %}
        // downscale signed integer
        if (in > {{ in.signed_type }}({{ in.signed_max }} - \
({{ in.signed_type }}(1) << {{ in.width - out.width - 1 }}))) {
            // clip
            out = {{ out.signed_max }};
        } else {
            out = {{ out.signed_type }}({{ in.unsigned_type }}(in + \
({{ in.signed_type }}(1) << {{ in.width - out.width - 1 }})) >> {{ in.width - out.width }});
        }
{% elif out.width > in.width and both_unsigned %}
        // upscale unsigned integer
        out = {{ out.type }}({{ out.type }}(in) << {{ out.width - in.width }});
{% elif out.width > in.width and not both_unsigned %}
        // upscale signed integer
        out = {{ out.signed_type }}({{ out.unsigned_type }}(in) << {{ out.width - in.width }});
{% endif %}

{% if not out.is_signed and not both_unsigned %}
        // convert to unsigned
        return pcm_sign_converter<PcmCode_{{ out.code }}>::from_signed(out);
{% else %}
        return out;
{% endif %}
{% endif %}
    }
};

{% endfor %}
{% endfor %}
// N-byte native-endian packed octet array
template <size_t N> struct pcm_octets;

{% for size in [1, 2, 4, 8] %}
// {{ size }}-byte native-endian packed octet array
template <> ROC_ATTR_PACKED_BEGIN struct pcm_octets<{{ size }}> {
#if ROC_CPU_ENDIAN == ROC_CPU_BE
{% for n in reversed(range(size)) %}
    uint8_t octet{{ n }};
{% endfor %}
#else
{% for n in range(size) %}
    uint8_t octet{{ n }};
{% endfor %}
#endif
} ROC_ATTR_PACKED_END;

{% endfor %}
// N-byte native-endian sample
template <class T> struct pcm_sample;

{% for type, size in TYPES %}
// {{ type }} native-endian sample
template <> struct pcm_sample<{{ type }}> {
    union {
        {{ type }} value;
        pcm_octets<{{ size }}> octets;
    };
};

{% endfor %}
// Write octet at given byte-aligned bit offset
inline void pcm_aligned_write(uint8_t* buffer, size_t& bit_offset, uint8_t arg) {
    buffer[bit_offset >> 3] = arg;
    bit_offset += 8;
}

// Read octet at given byte-aligned bit offset
inline uint8_t pcm_aligned_read(const uint8_t* buffer, size_t& bit_offset) {
    uint8_t ret = buffer[bit_offset >> 3];
    bit_offset += 8;
    return ret;
}

// Write value (at most 8 bits) at given unaligned bit offset
inline void
pcm_unaligned_write(uint8_t* buffer, size_t& bit_offset, size_t bit_length, uint8_t arg) {
    size_t byte_index = (bit_offset >> 3);
    size_t bit_index = (bit_offset & 0x7u);

    if (bit_index == 0) {
        buffer[byte_index] = 0;
    }

    buffer[byte_index] |= uint8_t(arg << (8 - bit_length) >> bit_index);

    if (bit_index + bit_length > 8) {
        buffer[byte_index + 1] = uint8_t(arg << bit_index);
    }

    bit_offset += bit_length;
}

// Read value (at most 8 bits) at given unaligned bit offset
inline uint8_t
pcm_unaligned_read(const uint8_t* buffer, size_t& bit_offset, size_t bit_length) {
    size_t byte_index = (bit_offset >> 3);
    size_t bit_index = (bit_offset & 0x7u);

    uint8_t ret = uint8_t(buffer[byte_index] << bit_index >> (8 - bit_length));

    if (bit_index + bit_length > 8) {
        ret |= uint8_t(buffer[byte_index + 1] >> (8 - bit_index) >> (8 - bit_length));
    }

    bit_offset += bit_length;
    return ret;
}

// Sample packer / unpacker
template <PcmCode, PcmEndian> struct pcm_packer;

{% for code in CODES %}
{% for endian in ['Big', 'Little'] %}
// {{ code.code }} {{ endian }}-Endian packer / unpacker
template <> struct pcm_packer<PcmCode_{{ code.code }}, PcmEndian_{{ endian }}> {
    // Pack next sample to buffer
    static inline void pack(uint8_t* buffer, size_t& bit_offset, {{ code.type }} arg) {
        // native-endian view of octets
        pcm_sample<{{ code.type }}> p;
        p.value = arg;

{% if code.width < code.packed_width %}
        // zeroise padding bits
        p.value &= {{ code.value_mask }};

{% endif %}
        // write in {{ endian.lower() }}-endian order
{% for n in range(code.packed_octets) %}
{% if endian == 'Big' %}
{% set n = code.packed_octets - n - 1 %}
{% endif %}
{% if code.packed_width % 8 == 0 %}
        pcm_aligned_write(buffer, bit_offset, p.octets.octet{{ n }});
{% else %}
        pcm_unaligned_write(buffer, bit_offset, \
{{ code.packed_width % 8 if n == code.packed_octets-1 else 8 }}, p.octets.octet{{ n }});
{% endif %}
{% endfor %}
    }

    // Unpack next sample from buffer
    static inline {{ code.type }} unpack(const uint8_t* buffer, size_t& bit_offset) {
        // native-endian view of octets
        pcm_sample<{{ code.type }}> p;

        // read in {{ endian.lower() }}-endian order
{% for n in range(code.unpacked_octets) %}
{% if endian == 'Big' %}
{% set n = code.unpacked_octets - n - 1 %}
{% endif %}
{% if n >= code.packed_octets %}
        p.octets.octet{{ n }} = 0;
{% elif code.packed_width % 8 == 0 %}
        p.octets.octet{{ n }} = pcm_aligned_read(buffer, bit_offset);
{% else %}
        p.octets.octet{{ n }} = pcm_unaligned_read(buffer, bit_offset, \
{{ code.packed_width % 8 if n == code.packed_octets-1 else 8 }});
{% endif %}
{% endfor %}

{% if code.width < code.packed_width %}
        // zeroise padding bits
        p.value &= {{ code.value_mask }};

{% endif %}
{% if code.is_signed and code.width < code.unpacked_width %}
        if (p.value & {{ code.sign_mask }}) {
            // sign extension
            p.value |= ({{ code.type }}){{ code.lsb_mask }};
        }

{% endif %}
        return p.value;
    }
};

{% endfor %}
{% endfor %}
// Mapping function implementation
template <PcmCode InCode, PcmEndian InEndian, PcmCode OutCode, PcmEndian OutEndian>
struct pcm_mapper {
    static void map(const uint8_t* in_data,
                    size_t& in_bit_off,
                    uint8_t* out_data,
                    size_t& out_bit_off,
                    size_t n_samples) {
        for (size_t n = 0; n < n_samples; n++) {
            pcm_packer<OutCode, OutEndian>::pack(
                out_data, out_bit_off,
                pcm_code_converter<InCode, OutCode>::convert(
                    pcm_packer<InCode, InEndian>::unpack(in_data, in_bit_off)));
        }
    }
};

// Select mapping function
template <PcmCode InCode, PcmEndian InEndian, PcmCode OutCode, PcmEndian OutEndian>
PcmMapFn pcm_format_mapfn() {
    return &pcm_mapper<InCode, InEndian, OutCode, OutEndian>::map;
}

// Select mapping function
template <PcmCode InCode, PcmEndian InEndian>
PcmMapFn pcm_format_mapfn(PcmFormat out_format) {
    switch (out_format) {
{% for code in CODES %}
{% for endian in ENDIANS %}
    case {{ make_enum_name(code, endian) }}:
{% if endian == 'Native' %}
#if ROC_CPU_ENDIAN == ROC_CPU_BE
        return pcm_format_mapfn<InCode, InEndian, PcmCode_{{ code.code }}, PcmEndian_Big>();
#else
        return pcm_format_mapfn<InCode, InEndian, PcmCode_{{ code.code }}, PcmEndian_Little>();
#endif
{% else %}
        return pcm_format_mapfn<InCode, InEndian, PcmCode_{{ code.code }}, PcmEndian_{{ endian }}>();
{% endif %}
{% endfor %}
{% endfor %}
    default:
        break;
    }
    return NULL;
}

} // namespace

// Select mapping function
PcmMapFn pcm_format_mapfn(PcmFormat in_format, PcmFormat out_format) {
    switch (in_format) {
{% for code in CODES %}
{% for endian in ENDIANS %}
    case {{ make_enum_name(code, endian) }}:
{% if endian == 'Native' %}
#if ROC_CPU_ENDIAN == ROC_CPU_BE
        return pcm_format_mapfn<PcmCode_{{ code.code }}, PcmEndian_Big>(out_format);
#else
        return pcm_format_mapfn<PcmCode_{{ code.code }}, PcmEndian_Little>(out_format);
#endif
{% else %}
        return pcm_format_mapfn<PcmCode_{{ code.code }}, PcmEndian_{{ endian }}>(out_format);
{% endif %}
{% endfor %}
{% endfor %}
    default:
        break;
    }
    return NULL;
}

// Get format traits
PcmTraits pcm_format_traits(PcmFormat format) {
    PcmTraits traits;

    switch (format) {
{% for code in CODES %}
{% for endian in ENDIANS %}
    case {{ make_enum_name(code, endian) }}:
        traits.is_valid = true;
        traits.is_integer = {{ str(code.is_integer).lower() }};
        traits.is_signed = {{ str(code.is_signed).lower() }};
{% if endian == 'Native' %}
#if ROC_CPU_ENDIAN == ROC_CPU_BE
        traits.is_little = false;
        traits.canon_id = {{ make_enum_name(code, 'Big') }};
#else
        traits.is_little = true;
        traits.canon_id = {{ make_enum_name(code, 'Little') }};
#endif
{% else %}
        traits.is_little = {{ str(endian == 'Little').lower() }};
        traits.canon_id = {{ make_enum_name(code, endian) }};
{% endif %}
        traits.bit_depth = {{ code.depth }};
        traits.bit_width = {{ code.packed_width }};
        break;

{% endfor %}
{% endfor %}
    default:
        break;
    }

    return traits;
}

const char* pcm_format_to_str(PcmFormat format) {
    switch (format) {
{% for code in CODES %}
{% for endian in ENDIANS %}
    case {{ make_enum_name(code, endian) }}:
        return "{{ make_str_name(code, endian) }}";
{% endfor %}
{% endfor %}
    default:
        break;
    }
    return NULL;
}

PcmFormat pcm_format_from_str(const char* str) {
    if (!str) {
        return PcmFormat_Invalid;
    }
{% for c0 in nth_chars(CODES) %}
    if (str[0] == '{{ c0 }}') {
{% for c1 in nth_chars(CODES, (c0,)) %}
        if (str[1] == '{{ c1 }}') {
{% for c2 in nth_chars(CODES, (c0, c1)) %}
            if (str[2] == '{{ c2 }}') {
{% for code in CODES %}
{% if tuple(code.short_name[:3]) == (c0, c1, c2) %}
{% for endian in ENDIANS %}
                if (strcmp(str, "{{ make_str_name(code, endian) }}") == 0) {
                    return {{ make_enum_name(code, endian) }};
                }
{% endfor %}
{% endif %}
{% endfor %}
                return PcmFormat_Invalid;
            }
{% endfor %}
{% for code in CODES %}
{% if tuple(code.short_name) == (c0, c1) %}
{% for endian in ENDIANS %}
            if (strcmp(str, "{{ make_str_name(code, endian) }}") == 0) {
                return {{ make_enum_name(code, endian) }};
            }
{% endfor %}
{% endif %}
{% endfor %}
            return PcmFormat_Invalid;
        }
{% endfor %}
        return PcmFormat_Invalid;
    }
{% endfor %}
    return PcmFormat_Invalid;
}

} // namespace audio
} // namespace roc
'''.strip())

text = template.render(
    **dict(list(globals().items()) + list(builtins.__dict__.items())),
    )

os.chdir(os.path.dirname(os.path.abspath(__file__)))

with open('pcm_format.cpp', 'w') as fp:
    print(text, file=fp)



import SCons.Script
import os
import re

def _is_test_enabled(kind, testname):
    for target in [kind, testname]:
        if target in SCons.Script.COMMAND_LINE_TARGETS:
            return True

def _get_non_test_targets(env):
    if SCons.Script.COMMAND_LINE_TARGETS:
        for target in SCons.Script.COMMAND_LINE_TARGETS:
            if target == 'test' or target == 'bench':
                yield env.Dir('#')
            elif not re.match('^(test|bench)/.+', target):
                yield target
    else:
        yield env.Dir('#')

def _run_with_timeout(env, cmd, timeout):
    return '{} scripts/scons_helpers/timeout-run.py {} {}'.format(
        env.GetPythonExecutable(),
        timeout,
        cmd)

def _add_test(env, kind, name, exe, cmd, timeout):
    varname = '_{}_TARGETS'.format(kind.upper())
    testname = '{}/{}'.format(kind, name)

    if not _is_test_enabled(kind, testname):
        return

    if not cmd:
        cmd = env.File(exe).path

    if timeout is not None:
        cmd = _run_with_timeout(env, cmd, timeout)

    if kind == 'test':
        cmd += ' -b'

    comstr = env.PrettyCommand(kind.upper(), name, 'green' if kind == 'test' else 'cyan')
    target = env.Alias(testname, [], env.Action(cmd, comstr))

    # This target produces no files.
    env.AlwaysBuild(target)

    # This target depends on test executable that it should run.
    env.Depends(target, env.File(exe))

    # This target should be run after all build targets.
    for t in _get_non_test_targets(env):
        env.Requires(target, t)

    # This target should be run after all previous tests.
    for t in env[varname]:
        env.Requires(target, t)

    # Benchmarks should be run after all tests.
    if kind == 'bench':
        for t in env['_TEST_TARGETS']:
            env.Requires(target, t)
    else:
        for t in env['_BENCH_TARGETS']:
            env.Requires(t, target)

    # Add target to test list.
    env[varname] += [testname]

    # 'test' target depends on this target.
    env.Depends(kind, target)

def AddTest(env, name, exe, cmd=None, timeout=5*60):
    _add_test(env, 'test', name, exe, cmd, timeout)

def AddBench(env, name, exe, cmd=None, timeout=None):
    _add_test(env, 'bench', name, exe, cmd, timeout)

def init(env):
    env['_TEST_TARGETS'] = []
    env['_BENCH_TARGETS'] = []

    env.AlwaysBuild(env.Alias('test', [], env.Action('')))
    env.AlwaysBuild(env.Alias('bench', [], env.Action('')))

    env.Requires('bench', 'test')

    env.AddMethod(AddTest, 'AddTest')
    env.AddMethod(AddBench, 'AddBench')



#!/usr/bin/env python3

import fileinput
import numpy
import pylab
import sys

text = []
columns = []

for line in fileinput.input(encoding='utf-8'):
    line = line.strip()
    if line.startswith('#'):
        columns = line.replace('#', '').split(',')
        continue
    text.append(line)

data = numpy.genfromtxt(text, dtype=float, delimiter=',')

if len(data.shape) == 1:
    data = data[:,numpy.newaxis]

if not columns:
    columns = [None] * data.shape[1]

fig, axs = pylab.subplots(len(columns), 1, sharex=True)

if len(columns) == 1:
    axs = [axs]

for col_idx, col_name in enumerate(columns):
    axs[col_idx].ticklabel_format(style='plain')
    if col_name:
        axs[col_idx].set_title(col_name)
    axs[col_idx].grid()
    axs[col_idx].plot(data[:,col_idx], '-*')

pylab.show()



import SCons.Script
import os
import os.path
import re
import shutil
import sys

try:
    from shlex import quote
except:
    from pipes import quote

def ClangFormat(env, src_dir):
    exclude_files = [
        os.path.relpath(env.File('#'+s).srcnode().abspath)
            for s in open(env.File('#.fmtignore').abspath).read().split()
        ]

    files = env.GlobRecursive(
        src_dir, ['*.h', '*.cpp'],
        exclude=exclude_files)

    return env.Action(
        '{clang_format} -i {files}'.format(
            clang_format=quote(env['CLANG_FORMAT']),
            files=' '.join(map(str, files))),
        env.PrettyCommand('FMT', env.Dir(src_dir).path, 'yellow'))

def HeaderFormat(env, src_dir):
    return env.Action(
        '{python} scripts/scons_helpers/format-header.py {src_dir}'.format(
            python=quote(env.GetPythonExecutable()),
            src_dir=quote(env.Dir(src_dir).path)),
        env.PrettyCommand('FMT', env.Dir(src_dir).path, 'yellow'))

def Doxygen(env, build_dir='', html_dir=None, config='', sources=[], werror=False):
    target = os.path.join(build_dir, 'commit')

    dirs = [env.Dir(build_dir).path]
    if html_dir:
        dirs += [env.Dir(html_dir).path]

    cmd = [
        quote(env.GetPythonExecutable()), 'scripts/scons_helpers/docfilt.py',
        '--root-dir', quote(env.Dir('#').path),
        '--work-dir', quote(env.Dir(os.path.dirname(config)).path),
        '--out-dirs', ' '.join(map(quote, dirs)),
        '--touch-file', quote(env.File(target).path),
        ]

    if werror:
        cmd += ['--werror']

    cmd += [
        '--',
        quote(env['DOXYGEN']),
        quote(env.File(config).name),
        ]

    env.Command(target, sources + [config], SCons.Action.CommandAction(
        ' '.join(cmd),
        cmdstr = env.PrettyCommand('DOXYGEN', env.Dir(build_dir).path, 'purple')))

    return target

def Sphinx(env, output_type, build_dir, output_dir, source_dir, sources, werror=False):
    target = os.path.join(build_dir, 'commit')

    cmd = [
        quote(env.GetPythonExecutable()), 'scripts/scons_helpers/docfilt.py',
        '--root-dir', quote(env.Dir('#').path),
        '--work-dir', quote(env.Dir('#').path),
        '--out-dirs', quote(env.Dir(output_dir).path),
        '--touch-file', quote(env.File(target).path),
        ]

    if werror:
        cmd += ['--werror']

    cmd += [
        '--',
        quote(env['SPHINX_BUILD']),
        '-j', str(SCons.Script.GetOption('num_jobs')),
        '-q',
        '-b', output_type,
        '-d', quote(env.Dir(build_dir).path),
        quote(env.Dir(source_dir).path),
        quote(env.Dir(output_dir).path),
        ]

    env.Command(target, sources, SCons.Action.CommandAction(
        ' '.join(cmd),
        cmdstr = env.PrettyCommand('SPHINX', env.Dir(output_dir).path, 'purple')))

    return env.File(target)

def Ragel(env, source, target=None):
    if 'RAGEL' in env.Dictionary():
        ragel = env['RAGEL']
    else:
        ragel = 'ragel'

    if not isinstance(ragel, str):
        ragel = env.File(ragel).path

    rl_file = env.File(source)

    cpp_file = env.File(os.path.join(
        str(source.dir),
        os.path.splitext(os.path.basename(source.path))[0] + '.cpp'))

    env.Command(cpp_file, rl_file, SCons.Action.CommandAction(
        '{ragel} -o {target} {source}'.format(
            ragel=quote(ragel),
            target=quote(cpp_file.path),
            source=quote(rl_file.srcnode().path)),
        cmdstr = env.PrettyCommand('RAGEL', '$SOURCE', 'purple')))

    return [env.Object(target=target, source=cpp_file)]

def GenGetOpt(env, source, version):
    if 'GENGETOPT' in env.Dictionary():
        gengetopt = env['GENGETOPT']
    else:
        gengetopt = 'gengetopt'

    if not isinstance(gengetopt, str):
        gengetopt = env.File(gengetopt).path

    source = env.File(source)
    source_name = os.path.splitext(os.path.basename(source.path))[0]
    target = [
        os.path.join(str(source.dir), source_name + '.c'),
        os.path.join(str(source.dir), source_name + '.h'),
    ]

    env.Command(target, source, SCons.Action.CommandAction(
        ('{gengetopt} -i {source} -F {source_name} --output-dir {output_dir}'
         ' --set-version {version}').format(
            gengetopt=quote(gengetopt),
            source=quote(source.srcnode().path),
            source_name=quote(source_name),
            output_dir=quote(os.path.dirname(source.path)),
            version=quote(version)),
        cmdstr = env.PrettyCommand('GGO', '$SOURCE', 'purple')))

    return env.Object(target[0])

def SupportsRelocatableObject(env):
    if not env.get('LD', None):
        return False

    out = env.GetCommandOutput('{} -V'.format(env['LD']))
    return 'GNU' in out

def RelocatableObject(env, dst, src_list):
    dst += env['OBJSUFFIX']

    action = SCons.Action.CommandAction(
        '$LD -r $SOURCES -o $TARGET',
        cmdstr=env.PrettyCommand('LD', env.File(dst).path, 'red'))

    return env.Command(dst, src_list, [action])

def SupportsLocalizedObject(env):
    if not env.get('OBJCOPY', None):
        return False

    out = env.GetCommandOutput('{} -V'.format(env['OBJCOPY']))
    return 'GNU' in out

def LocalizedObject(env, dst, src):
    dst += env['OBJSUFFIX']

    action = SCons.Action.CommandAction(
        '$OBJCOPY --localize-hidden --strip-unneeded $SOURCE $TARGET',
        cmdstr=env.PrettyCommand('OBJCOPY', env.File(dst).path, 'red'))

    return env.Command(dst, src, [action])

def SupportsStripSharedLibrary(env):
    return env.get('STRIP', None) is not None

def StripSharedLibrary(env, dst, src):
    def _copy(target, source, env):
        shutil.copy(source[0].path, target[0].path)

    actions =  [
        env.Action(
            _copy,
            env.PrettyCommand('CP', env.File(dst).path, 'yellow'),
            ),
        SCons.Action.CommandAction(
            '$STRIP $STRIPFLAGS $TARGET',
            cmdstr=env.PrettyCommand('STRIP', '$TARGET', 'red'),
            ),
    ]

    return env.Command(dst, src, actions)

def SymlinkLibrary(env, src):
    def _symlink(target, source, env):
        os.symlink(os.path.relpath(source[0].path, os.path.dirname(target[0].path)),
                   target[0].path)

    path = src.abspath
    ret = []

    while True:
        m = re.match(r'^(.+)\.[0-9]+(\.[a-z]+)?$', path)
        if not m:
            break

        path = m.group(1) + (m.group(2) or '')

        dst = env.File(path)
        ret += [dst]

        env.Command(dst, src, env.Action(
            _symlink, env.PrettyCommand('LN', dst.path, 'yellow', 'ln({})'.format(dst.path))))

    return ret

def NeedsFixupSharedLibrary(env):
    return env.get('INSTALL_NAME_TOOL', None) is not None

def FixupSharedLibrary(env, path):
    return [
        SCons.Action.CommandAction(
            '$INSTALL_NAME_TOOL -id {0} {0}'.format(quote(path)),
            cmdstr=env.PrettyCommand('FIXUP', path, 'yellow')),
            ]

def ComposeStaticLibraries(env, dst_lib, src_libs):
    dst_lib = env['LIBPREFIX'] + dst_lib + env['LIBSUFFIX']

    cmd = [
        quote(env.GetPythonExecutable()), 'scripts/scons_helpers/compose-libs.py',
        '--out', quote(env.File(dst_lib).path),
        '--in', ' '.join([quote(env.File(lib).path) for lib in src_libs]),
        ]

    if env['ROC_PLATFORM'] == 'darwin' and env['ROC_MACOS_ARCH']:
        cmd += ['--arch', ' '.join(env['ROC_MACOS_ARCH'])]

    cmd += ['--tools']
    for tool in ['AR', 'OBJCOPY', 'LIPO']:
        if env.get(tool, None):
            cmd += [
                '{}={}'.format(tool, quote(env[tool])),
                ]

    action = SCons.Action.CommandAction(
        ' '.join(cmd),
        cmdstr=env.PrettyCommand('COMPOSE', env.File(dst_lib).path, 'red'))

    return env.Command(dst_lib, [src_libs[0]], [action])

def DeleteFile(env, path):
    path = env.File(path).path

    def rmfile(target, source, env):
        if os.path.exists(path):
            os.remove(path)

    return env.Action(rmfile, env.PrettyCommand('RM', path, 'red', 'rm({})'.format(path)))

def DeleteDir(env, path):
    path = env.Dir(path).path

    def rmtree(target, source, env):
        if os.path.exists(path):
            shutil.rmtree(path)

    return env.Action(rmtree, env.PrettyCommand('RM', path, 'red', 'rm({})'.format(path)))

def Artifact(env, dst, src):
    def noop(target, source, env):
        pass

    target = env.File(dst)

    env.Command(dst, src, env.Action(noop, env.PrettyCommand(
        'ART', target.path, 'yellow', 'art({})'.format(target.path))))

    env.Precious(dst)
    env.Requires(dst, src)

    return target

def init(env):
    env.AddMethod(ClangFormat, 'ClangFormat')
    env.AddMethod(HeaderFormat, 'HeaderFormat')
    env.AddMethod(Doxygen, 'Doxygen')
    env.AddMethod(Sphinx, 'Sphinx')
    env.AddMethod(Ragel, 'Ragel')
    env.AddMethod(GenGetOpt, 'GenGetOpt')
    env.AddMethod(SupportsRelocatableObject, 'SupportsRelocatableObject')
    env.AddMethod(RelocatableObject, 'RelocatableObject')
    env.AddMethod(SupportsLocalizedObject, 'SupportsLocalizedObject')
    env.AddMethod(LocalizedObject, 'LocalizedObject')
    env.AddMethod(SupportsStripSharedLibrary, 'SupportsStripSharedLibrary')
    env.AddMethod(StripSharedLibrary, 'StripSharedLibrary')
    env.AddMethod(SymlinkLibrary, 'SymlinkLibrary')
    env.AddMethod(NeedsFixupSharedLibrary, 'NeedsFixupSharedLibrary')
    env.AddMethod(FixupSharedLibrary, 'FixupSharedLibrary')
    env.AddMethod(ComposeStaticLibraries, 'ComposeStaticLibraries')
    env.AddMethod(DeleteFile, 'DeleteFile')
    env.AddMethod(DeleteDir, 'DeleteDir')
    env.AddMethod(Artifact, 'Artifact')



#! /usr/bin/env python3

import datetime
import fileinput
import os
import os.path
import re
import subprocess
import sys
import textwrap

os.chdir(os.path.join(
    os.path.dirname(__file__), '..'))

def update_deb_copyright():
    in_path = 'docs/sphinx/about_project/authors.rst'
    out_path = 'debian/copyright'

    print(f'updating {out_path}')

    with open(in_path, 'r') as in_file:
        in_lines = in_file.readlines()

    with open(out_path, 'r') as out_file:
        out_content = out_file.readlines()

    updated_content = []

    for line in out_content:
        updated_content.append(line)
        if line.startswith('Copyright:'):
            break

    for line in in_lines:
        if line.startswith('* '):
            updated_content.append(line.replace('* ', '  '))

    for line in out_content:
        if line.startswith('License:'):
            updated_content.append(line)
            break

    with open(out_path, 'w') as out_file:
        out_file.writelines(updated_content)

def update_deb_changelog():
    in_path = 'docs/sphinx/development/changelog.rst'
    out_path = 'debian/changelog'

    print(f'updating {out_path}')

    in_file = open(in_path, 'r')
    out_file = open(out_path, 'w')

    def fprint(msg=''):
        print(msg, file=out_file)

    first = True

    for line in in_file.readlines():
        pad = re.sub(r'^(\s*).*$', r'\1', line.rstrip())
        line = line.strip()

        if line.startswith('==='):
            if not first:
                fprint()
                fprint(f' -- Roc Streaming authors <roc@freelists.org>  {prev_date}')
                fprint()
            first = False

            m = re.match('^Version\s+(\S+)\s+\((.*)\)$', prev_line)
            version, date = m.group(1), m.group(2)
            prev_date = datetime.datetime.strptime(date, "%b %d, %Y").\
                replace(tzinfo=datetime.timezone.utc).\
                strftime("%a, %d %b %Y %H:%M:%S %z")
            fprint(f'roc-toolkit ({version}) unstable; urgency=low')

        if line.startswith('---'):
            fprint()
            fprint(f'  [ {prev_line} ]')

        if line.startswith('* '):
            line = re.sub(r'`([^` ]+)\s*<[^` >]+>`_+', r'\1', line)
            line = re.sub(r'``', '`', line)
            fprint('\n'.join(textwrap.wrap(
                line, width=80, initial_indent=('  '+pad), subsequent_indent=('    '+pad))))

        prev_line = line

    fprint()
    fprint(f' -- Roc Streaming authors <roc@freelists.org>  {prev_date}')
    fprint()

def update_rpm_spec():
    out_path = 'rpm/roc-toolkit.spec'

    print(f'updating {out_path}')

    version = subprocess.check_output([
        sys.executable, 'scripts/scons_helpers/parse-version.py']).decode().strip()

    for line in fileinput.input(files=[out_path], inplace=True):
        m = re.match('^(Version:\s+).*$', line)
        if m:
            line = m.group(1) + version + '\n'
        print(line, end='')

update_deb_copyright()
update_deb_changelog()
update_rpm_spec()

print('done')
